# -*- coding: utf-8 -*-
"""recipe cooking"""

import json
import yaml
import re
import os
from orderedattrdict import AttrDict
from orderedattrdict.yamlutils import AttrDictYAMLLoader

from . helpers import read_opt, mkfunc
from functools import reduce
from .. import config
from . exceptions import ChefRuntimeError
from ..model.ddf import Dataset
from . import procedure

import logging

logger = logging.getLogger('Chef')


class Pipeline:
    def __init__(self, name, procedures, chef=None, result=None):
        self.name = name
        self.procedures = procedures
        self._result = result
        self.chef = chef

    @property
    def is_cooked(self):
        if self._result is not None:
            return True
        else:
            return False

    @property
    def result(self):
        if self._result is not None:
            return self._result
        else:
            self.cook()
            return self._result

    def cook(self):
        def get_proc_func(proc):
            func = getattr(procedure, proc['procedure'])
            return func

        def run_procedure(proc):
            func = get_proc_func(proc)
            return func(self.chef, **proc['options'])

        for proc in self.procedures:
            self._result = run_procedure(proc)

        return self._result


class Chef:
    def __init__(self, info=None, cfg=None, ingredients=None, cooking=None, serving=None):
        assert 'main' in cooking.keys(), 'the cooking pipeline "main" must exists'

        self.info = info
        self.config = cfg
        self._ingredients = ingredients
        self.cooking = cooking
        self.serving = serving
        self._ingredients_bag = None

    @classmethod
    def from_recipe(cls, recipe):
        info = recipe['info']
        cfg = recipe['config']
        ingredients = recipe['ingredients']
        cooking = recipe['cooking']
        serving = recipe['serving']

        return cls(info, cfg, ingredients, cooking, serving)

    @property
    def ingredients_bag(self):
        if self._ingredients_bag is not None:
            return self._ingredients_bag
        bag = dict()
        for ing in self._ingredients:
            bag[ing['id']] = Dataset.from_ddfcsv(ing['path'])
        for name, ppl in self.cooking.items():
            bag[name] = Pipeline(name, ppl, self)
        self._ingredients_bag = bag
        return bag

    def get_ingredient(self, name):
        ing = self.ingredients_bag[name]
        if isinstance(ing, Pipeline):
            return ing.result
        else:
            return ing

    def cook(self):
        result = self.get_ingredient('main').compute()
        result.serve(self.serving)


def _loadfile(f):
    """load json/yaml file, into AttrDict"""
    if re.match('.*\.json', f):
        res = json.load(open(f), object_pairs_hook=AttrDict)
    else:
        res = yaml.load(open(f), Loader=AttrDictYAMLLoader)

    return res





# -*- coding: utf-8 -*-

"""run_op procedure for recipes"""

import logging
from typing import List

import pandas as pd

from .. helpers import debuggable
from .. model.ingredient import DataPointIngredient
from .. model.chef import Chef


logger = logging.getLogger('run_op')


@debuggable
def run_op(chef: Chef, ingredients: List[DataPointIngredient], result, op) -> DataPointIngredient:
    """run math operation on each row of ingredient data.

    Procedure format:

    .. code-block:: yaml

       procedure: run_op
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingredient id
       options:
         op: dict  # a dictionary describing calculation for each columns.

    Keyword Args
    ------------
    op: dict
        a dictionary of concept_name -> function mapping

    Examples
    --------
    .. highlight:: yaml

    for exmaple, if we want to add 2 columns, col_a and col_b, to create an new column, we can
    write

    ::

        procedure: run_op
        ingredients:
          - ingredient_to_run
        result: new_ingredient_id
        options:
          op:
            new_col_name: "col_a + col_b"
    """

    assert len(ingredients) == 1, "procedure only support 1 ingredient for now."
    # ingredient = chef.dag.get_node(ingredients[0]).evaluate()
    ingredient = ingredients[0]
    assert ingredient.dtype == 'datapoints'
    logger.info("run_op: " + ingredient.id)

    # FIXME: avoid using compute
    data = ingredient.compute()
    keys = ingredient.key

    # concat all the datapoint dataframe first, and eval the ops
    to_concat = [v for v in data.values()]

    # preserve dtypes
    dtypes_ = to_concat[0].dtypes[keys]
    dtypes = dict([k, v.name] for k, v in dtypes_.items())

    if len(to_concat) == 1:
        df = to_concat[0]
    else:
        df = pd.merge(to_concat[0], to_concat[1], on=keys, how='outer')
        for _df in to_concat[2:]:
            df = pd.merge(df, _df, on=keys, how='outer')
        # reset dtypes
        for c in keys:
            df[c] = df[c].astype(dtypes[c])
    df = df.set_index(keys)

    for k, v in op.items():
        res = df.eval(v).dropna()  # type(res) is Series
        res.name = k
        if k not in df.columns:
            df[k] = res
        data[k] = res.reset_index()

    if not result:
        result = ingredient.id + '-op'
    return DataPointIngredient.from_procedure_result(result, ingredient.key, data_computed=data)



# -*- coding: utf-8 -*-

"""translate_column procedures for recipes"""

import logging
from typing import List

from .. helpers import debuggable, build_dictionary
from .. model.ingredient import Ingredient, get_ingredient_class
from .. model.chef import Chef

logger = logging.getLogger('translate_column')


@debuggable
def translate_column(chef: Chef, ingredients: List[Ingredient], result, dictionary,
                     column, *, target_column=None, not_found='drop',
                     ambiguity='prompt', ignore_case=False) -> Ingredient:
    """Translate column values.

    Procedure format:

    .. code-block:: yaml

       procedure: translate_column
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingredient id
       options:
         column: str  # the column to be translated
         target_column: str  # optional, the target column to store the translated data
         not_found: {'drop', 'include', 'error'}  # optional, the behavior when there is values not
                                                  # found in the mapping dictionary, default is 'drop'
         ambiguity: {'prompt', 'skip', 'error'}  # optional, the behavior when there is ambiguity
                                                 # in the dictionary
         dictionary: str or dict  # file name or mappings dictionary

    If base is provided in dictionary, key and value should also in dictionary.
    In this case chef will generate a mapping dictionary using the base ingredient.
    The dictionary format will be:

    .. code-block:: yaml

       dictionary:
         base: str  # ingredient name
         key: str or list  # the columns to be the keys of the dictionary, can accept a list
         value: str  # the column to be the values of the the dictionary, must be one column

    Parameters
    ----------
    chef : Chef
        The Chef the procedure will run on
    ingredients : list
        A list of ingredient id in the dag to translate

    Keyword Args
    ------------
    dictionary: dict
        A dictionary of oldname -> newname mappings.
        If 'base' is provided in the dictionary, 'key' and 'value' should also in the dictionary.
        See :py:func:`ddf_utils.transformer.translate_column` for more on how this is handled.
    column: `str`
        the column to be translated
    target_column : `str`, optional
        the target column to store the translated data. If this is not set then the `column`
        column will be replaced
    not_found : {'drop', 'include', 'error'}, optional
        the behavior when there is values not found in the mapping dictionary, default is 'drop'
    ambiguity : {'prompt', 'skip', 'error'}, optional
        the behavior when there is ambiguity in the dictionary, default is 'prompt'

    See Also
    --------
    :py:func:`ddf_utils.transformer.translate_column` : related function in transformer module
    """
    from ... transformer import translate_column as tc
    assert len(ingredients) == 1, "procedure only support 1 ingredient for now."

    # ingredient = chef.dag.get_node(ingredients[0]).evaluate()
    ingredient = ingredients[0]
    logger.info("translate_column: " + ingredient.id)

    if target_column is None:
        target_column = column

    di = ingredient.get_data()
    new_data = dict()

    # build the dictionary
    dictionary_ = build_dictionary(chef, dictionary, ignore_case)
    dict_type = 'inline'
    base_df = None

    for k, df in di.items():
        logger.debug("running on: " + k)
        if df.dtypes[column].name == "category":
            categorical_col = True
        else:
            categorical_col = False

        if categorical_col:
            df[column] = df[column].astype('str')
        new_data[k] = tc(df, column, dict_type, dictionary_, target_column, base_df,
                         not_found, ambiguity, ignore_case)
        # TODO: astype() returns unknown category, but it's better to be known.
        if categorical_col:
            new_data[k][target_column] = new_data[k][target_column].astype("category")

    if not result:
        result = ingredient.id + '-translated'
    return get_ingredient_class(ingredient.dtype).from_procedure_result(
        result, ingredient.key, data_computed=new_data)



# -*- coding: utf-8 -*-

"""the DAG model of chef

The DAG consists of 2 types of nodes: IngredientNode and ProcedureNode.
each node will have a `evaluate()` function, which will return an ingredient
on eval.
"""

from . import procedure as pc
from .exceptions import ProcedureError, ChefRuntimeError


class BaseNode():
    """The base node which IngredientNode and ProcedureNode inherit from

    Parameters
    ----------
    node_id : `str`
        the name of the node
    dag : DAG
        the `DAG` object the node is in
    """
    def __init__(self, node_id, dag):
        self.node_id = node_id
        self.dag = dag
        self._upstream_list = list()
        self._downstream_list = list()

    def __repr__(self):
        return '<Node {}>'.format(self.node_id)

    @property
    def upstream_list(self):
        return [self.dag.get_node(nid) for nid in self._upstream_list]

    @property
    def downstream_list(self):
        return [self.dag.get_node(nid) for nid in self._downstream_list]

    def add_upstream(self, node):
        self._upstream_list.append(node.node_id)

    def add_downstream(self, node):
        self._downstream_list.append(node.node_id)

    def evaluate(self):
        raise NotImplementedError('')

    def get_direct_relatives(self, upstream=False):
        """
        Get the direct relatives to the current node, upstream or
        downstream.
        """
        if upstream:
            return self.upstream_list
        else:
            return self.downstream_list

    def detect_downstream_cycle(self, node=None):
        """
        When invoked, this routine will raise an exception if a cycle is
        detected downstream from self. It is invoked when nodes are added to
        the DAG to detect cycles.
        """
        if not node:
            node = self
        for t in self.get_direct_relatives():
            if node is t:
                msg = "Cycle detected in DAG. Faulty node: {0}".format(node)
                raise ChefRuntimeError(msg)
            else:
                t.detect_downstream_cycle(node=node)
        return False

    def detect_missing_dependency(self):
        """
        check if every upstream is available in the DAG.
        raise error if something is missing
        """
        not_found = set()
        for n in self.upstream_list:
            if not self.dag.has_node(n.node_id):
                not_found.add(n.node_id)
            if isinstance(n, ProcedureNode) and not n.procedure:
                not_found.add(n.node_id)
        if len(not_found) > 0:
            raise ChefRuntimeError(
                "dependency not found/not definded for {}: {}".format(self.node_id, not_found))
        return False


class IngredientNode(BaseNode):
    """Node for storing dataset ingredients.

    Parameters
    ----------
    ingredient : Ingredient
        the ingredient in this node
    """
    def __init__(self, node_id, ingredient, dag):
        super(IngredientNode, self).__init__(node_id, dag)
        self.ingredient = ingredient

    def evaluate(self):
        """return the ingredient as is"""
        return self.ingredient


class ProcedureNode(BaseNode):
    """The node for storing procedure results

    The evaluate() function will run a procedure according to `self.procedure`, using
    other nodes' data. Other nodes will be evaluated if when necessary.

    Parameters
    ----------
    procedure : dict
        the procedure dictionary
    """
    def __init__(self, node_id, procedure, dag):
        super(ProcedureNode, self).__init__(node_id, dag)
        self.procedure = procedure
        self.result_ingredient = None

    def evaluate(self):
        if self.result_ingredient:
            return self.result_ingredient

        # get the procedure function, raise error if procedure not supported
        try:
            func = getattr(pc, self.procedure['procedure'])
        except AttributeError:
            raise ProcedureError("Not supported: " + self.procedure['procedure'])
        except TypeError:
            raise ProcedureError("Procedure Error: " + str(self.node_id))

        # check the base ingredients and convert the string id to actual ingredient
        ingredients = []
        for i in self.procedure['ingredients']:
            ing = self.dag.get_node(i)
            ingredients.append(ing.evaluate())

        # also evaluate the ingredients in the options
        if 'options' in self.procedure.keys():
            options = self.procedure['options']
            for ingredient_key in ['base', 'ingredient']:
                if ingredient_key in options.keys():
                    ing = self.dag.get_node(self.procedure['options'][ingredient_key])
                    options[ingredient_key] = ing.evaluate()
                for opt in options.keys():
                    if isinstance(options[opt], dict):
                        if ingredient_key in options[opt].keys():
                            ing = self.dag.get_node(options[opt][ingredient_key])
                            options[opt][ingredient_key] = ing.evaluate()
            self.result_ingredient = func(*ingredients, result=self.procedure['result'], **options)
        else:
            self.result_ingredient = func(*ingredients, result=self.procedure['result'])

        return self.result_ingredient


class DAG():
    """The DAG model.

    A dag (directed acyclic graph) is a collection of tasks with directional
    dependencies. DAGs essentially act as namespaces for its nodes. A node_id
    can only be added once to a DAG.
    """
    def __init__(self, node_dict=None):
        if not node_dict:
            self._node_dict = dict()
        else:
            self._node_dict = node_dict

    @property
    def roots(self):
        """return the roots of the DAG"""
        return [t for t in self.nodes if not t.downstream_list]

    @property
    def nodes(self):
        """return all nodes"""
        return list(self.node_dict.values())

    @property
    def node_dict(self):
        return self._node_dict

    @node_dict.setter
    def node_dict(self, node):
        raise AttributeError('can not set node_dict manually')

    def add_node(self, node):
        """add a node to DAG"""
        if node.node_id in self.node_dict.keys():
            # only overwirte case is when procedure in ProcedureNode is None.
            if (isinstance(node, ProcedureNode) and
                not self.node_dict[node.node_id].procedure):
                self.node_dict[node.node_id] = node
            else:
                raise ChefRuntimeError('can not overwirte node already exists: ' + node.node_id)
        self.node_dict[node.node_id] = node

    def get_node(self, node_id):
        if node_id in self.node_dict.keys():
            return self.node_dict[node_id]
        raise ChefRuntimeError('node {} not found'.format(node_id))

    def has_node(self, node_id):
        return node_id in self.node_dict.keys()

    def add_dependency(self, upstream_node_id, downstream_node_id):
        """
        Simple utility method to set dependency between two nodes that
        already have been added to the DAG using add_node()
        """
        self.get_node(upstream_node_id).add_downstream(
            self.get_node(downstream_node_id))
        self.get_node(downstream_node_id).add_upstream(
            self.get_node(upstream_node_id))

    def tree_view(self):
        """
        Shows an ascii tree representation of the DAG
        """
        def get_downstream(node, level=0):
            print((" " * level * 4) + str(node))
            level += 1
            for t in node.upstream_list:
                get_downstream(t, level)

        for t in self.roots:
            get_downstream(t)



# -*- coding: utf-8 -*-

import attr
from abc import abstractmethod, ABC

import requests as req
from time import sleep
from functools import wraps
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry


# from https://www.peterbe.com/plog/best-practice-with-retries-with-requests
def requests_retry_session(
    retries=5,
    backoff_factor=0.3,
    status_forcelist=(500, 502, 504),
    session=None,
):
    session = session or req.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session


def retry(times=5, backoff=0.5):
    """general wrapper to retry things"""
    def wrapper(func):
        @wraps(func)
        def newfunc(*args, **kwargs):
            mtimes = times
            ttimes = times
            while ttimes > 0:
                try:
                    func(*args, **kwargs)
                except KeyboardInterrupt:
                    raise
                except Exception as e:
                    ttimes = ttimes - 1
                    sleep(backoff * (mtimes - ttimes))
                else:
                    break
        return newfunc
    return wrapper


@attr.s
class DataFactory(ABC):
    metadata = attr.ib(init=False, default=None)

    @abstractmethod
    def load_metadata(self, *args, **kwargs):
        ...

    @abstractmethod
    def has_newer_source(self, *args, **kwargs):
        ...

    @abstractmethod
    def bulk_download(self, *args, **kwargs):
        ...



# -*- coding: utf-8 -*-
"""string functions for ddf files"""

import re
import pandas as pd
from unidecode import unidecode
import decimal


def to_concept_id(s, sep='_'):
    """convert a string to alphanumeric format."""
    if pd.isnull(s):
        return s
    # replace some symbol to meaningful strings
    sub = {'%': 'pct', '>': 'gt', '<': 'lt'}
    for k, v in sub.items():
        if k in s:
            s = s.replace(k, v)

    pattern = re.compile('[\W_]+')  # match all Non-alphanumeric
    snew = unidecode(s.strip())
    snew = pattern.sub(sep, snew).lower()

    if len(snew.replace(sep, '')) == 0:
        return None

    # remove first/last underscore
    if snew[-1] == sep:
        snew = snew[:-1]
    if snew[0] == sep:
        snew = snew[1:]

    return snew


def fix_time_range(s):
    """change a time range to the middle of year in the range.
    e.g. fix_time_range('1980-90') = 1985
    """

    if '-' not in s:
        return int(s)
    else:
        t1, t2 = s.split('-')
        if len(t1) == 4 and len(t2) == 4:
            span = int(t2) - int(t1)
            return int(int(t1) + span / 2)
        else:  # t2 have only 1-2 digits.
            d = len(t2)
            hund1 = int(t1[:4 - d])
            tens1 = int(t1[-d:])
            tens2 = int(t2)
            y1 = int(t1)
            if tens1 > tens2:
                hund2 = hund1 + 1
                y2 = hund2 * 10**d + tens2
            else:
                y2 = hund1 * 10**d + tens2

            return int(y1 + (y2 - y1) / 2)


def _float_to_decimal(f):
    """Convert a floating point number to a Decimal with no loss of information

    see http://docs.python.org/library/decimal.html#decimal-faq
    """
    n, d = f.as_integer_ratio()
    numerator, denominator = decimal.Decimal(n), decimal.Decimal(d)
    ctx = decimal.Context(prec=60)
    result = ctx.divide(numerator, denominator)
    while ctx.flags[decimal.Inexact]:
        ctx.flags[decimal.Inexact] = False
        ctx.prec *= 2
        result = ctx.divide(numerator, denominator)
    return result


def format_float_digits(number, digits=5, threshold=None, keep_decimal=False):
    """format the number to string, limit the maximum amount of digits. Removing tailing zeros."""
    # assert(digits > 0)
    if pd.isnull(number):
        return number
    try:
        d = decimal.Decimal(number)
    except TypeError:
        d = _float_to_decimal(float(number))

    if threshold:
        if abs(d) <= threshold:
            return '0'

    s = format(d, '.{}f'.format(digits))

    if '.' in s:
        s = s.rstrip('0')
        if s[-1] == '.':
            if keep_decimal:
                s = s + '0'  # keep the decimal point and one zero.
            else:
                s = s[:-1]
            if s.startswith("-0"):
                s = "0"
    return s


def format_float_sigfig(number, sigfig=5, threshold=None):
    """format the number to string, keeping some significant digits."""
    # http://stackoverflow.com/questions/2663612/nicely-representing-a-floating-point-number-in-python/2663623#2663623
    # assert(sigfig > 0)
    if pd.isnull(number):
        return number
    try:
        d = decimal.Decimal(number)
    except TypeError:
        d = _float_to_decimal(float(number))

    if threshold:
        if abs(d) <= threshold:
            d = decimal.Decimal(0)

    sign, digits, exponent = d.as_tuple()

    if len(digits) < sigfig:
        digits = list(digits)
        digits.extend([0] * (sigfig - len(digits)))
    shift = d.adjusted()
    result = int(''.join(map(str, digits[:sigfig])))
    # Round the result
    if len(digits) > sigfig and digits[sigfig] >= 5:
        result += 1
    result = list(str(result))
    # Rounding can change the length of result
    # If so, adjust shift
    shift += len(result) - sigfig
    # reset len of result to sigfig
    result = result[:sigfig]
    if shift >= sigfig - 1:
        # Tack more zeros on the end
        result += ['0'] * (shift - sigfig + 1)
    elif 0 <= shift:
        # Place the decimal point in between digits
        result.insert(shift + 1, '.')
    else:
        # Tack zeros on the front
        assert (shift < 0)
        result = ['0.'] + ['0'] * (-shift - 1) + result
    if sign:
        result.insert(0, '-')
    return ''.join(result)



# -*- coding: utf-8 -*-

"""The Chef object"""

import json
import logging
import os
import re
import sys
from collections.abc import Mapping
from copy import deepcopy
from time import time

from dask import delayed

import ruamel.yaml as yaml
from graphviz import Digraph

from typing import List

from . dag import DAG, IngredientNode, ProcedureNode
from .. exceptions import ChefRuntimeError
from .. helpers import get_procedure, gen_sym, query, read_local_ddf
from . ingredient import Ingredient, ingredient_from_dict


logger = logging.getLogger(__name__)


def _loadfile(f):
    """load json/yaml file, into AttrDict"""
    if re.match(r'.*\.json', f):
        res = json.load(open(f))
    else:
        res = yaml.load(open(f), Loader=yaml.Loader)

    return res


# TODO: type annotate the class
class Chef:
    """the chef api"""

    def __init__(self, dag: DAG = None, metadata=None, config=None, cooking=None, serving=None, recipe=None):
        if dag is None:
            self.dag = DAG()
        else:
            self.dag = dag
        if metadata is None:
            self.metadata = dict()
        else:
            self.metadata = metadata
        if config is None:
            self._config = dict()
        else:
            self._config = config
        if cooking is None:
            self.cooking = {'concepts': list(), 'datapoints': list(), 'entities': list()}
        else:
            self.cooking = cooking
        if serving is None:
            self._serving = list()
        else:
            self._serving = serving

        self._recipe = recipe
        # self.ddf_object_cache = {}

    @property
    def config(self):
        if 'ddf_dir' not in self._config.keys():
            logger.warning('ddf_dir not configured, assuming current directory')
            self._config['ddf_dir'] = os.path.abspath('./')
        return self._config

    @property
    def serving(self):
        if len(self._serving) == 0:
            for k, v in self.cooking.items():
                if len(v) > 0:
                    self._serving.append({'id': v[-1]['result']})
        return self._serving

    @serving.setter
    def serving(self, serving):
        self._serving = serving

    @classmethod
    def from_recipe(cls, recipe_file, **config):
        recipe = Chef._build_recipe(recipe_file)
        chef = cls()

        if 'info' in recipe.keys():
            chef.add_metadata(**recipe['info'])
        if 'config' in recipe.keys():
            chef.add_config(**recipe['config'])
        chef.add_config(**config)

        assert 'ingredients' in recipe.keys(), "recipe must have ingredients section"
        assert 'cooking' in recipe.keys(), "recipe must have cooking section!"

        for ingred in recipe['ingredients']:
            chef.add_ingredient(**ingred)

        for k, ps in recipe['cooking'].items():
            for p in ps:
                result = p.get('result', None)
                options = p.get('options', None)
                chef.add_procedure(k, p['procedure'], p['ingredients'], result, options)

        chef.serving = Chef._get_dishes(recipe)
        return chef

    @property
    def ingredients(self) -> List[Ingredient]:
        return [x.evaluate() for x in self.dag.nodes if isinstance(x, IngredientNode)]

    def copy(self):
        return Chef(dag=self.dag.copy(), metadata=deepcopy(self.metadata),
                    config=deepcopy(self._config), cooking=deepcopy(self.cooking),
                    serving=deepcopy(self._serving), recipe=deepcopy(self._recipe))

    def validate(self):
        """validate if the chef is good to run.

        The following will be tested:

        1. check if datasets required by ingredients are available
        2. check if procedures are available
        3. check if the DAG is valid. i.e no dependency cycle, no missing dependency.
        """
        # 1. check dataset availability
        ddf_dir = self.config['ddf_dir']
        datasets = list()
        for ingred in self.ingredients:
            if ingred.ingredient_type == 'ddf':
                datasets.append(ingred)
        not_exists = set()
        for d in datasets:
            if not os.path.exists(d.dataset_path):
                not_exists.add(d.dataset)
        if len(not_exists) > 0:
            logger.critical("not enough datasets! please checkout following datasets:\n{}\n"
                            .format('\n'.join(list(not_exists))))
            raise ChefRuntimeError('not enough datasets')

        # 2. check procedure availability
        for k, ps in self.cooking.items():
            for p in ps:
                try:
                    get_procedure(p['procedure'], self.config.get('procedure_dir'))
                except (AttributeError, ImportError):
                    logger.warning("{} is not a valid procedure, please double check "
                                   "or register new procedure".format(p['procedure']))
                    raise ChefRuntimeError('procedures not ready')

        # 3. check if the DAG is valid
        cycle = self.dag.detect_cycles()
        if cycle:
            raise ChefRuntimeError('cycle detected: {}'.format(cycle))
        for ing in self.serving:
            self.dag.get_node(ing['id']).detect_missing_dependency()

    def add_config(self, **config):
        """add configs, all keyword args will be added/replace existing in config dictionary"""
        for k, v in config.items():
            self._config[k] = v
        return self

    def add_metadata(self, **metadata):
        """add metadata, all keyword args will be added/replace existing in metadata dictionary"""
        for k, v in metadata.items():
            self.metadata[k] = v
        return self

    def add_ingredient(self, **kwargs):
        """add a new ingredient in DAG.

        keyword arguments will send as a dictionary to the
        ``dictionary`` keyword of
        :py:func:`ddf_utils.chef.model.ingredient.ingredient_from_dict`
        method.

        """
        ingredient = ingredient_from_dict(dictionary=kwargs, **self.config)
        self.dag.add_node(IngredientNode(ingredient.id, ingredient, self))
        return self

    def add_procedure(self, collection, procedure, ingredients, result=None, options=None):

        if procedure == 'serve':
            [self._serving.append({'id': x,
                                   'options': options}) for x in ingredients]
            return self

        # check if procedure is supported
        try:
            get_procedure(procedure, self.config.get('procedure_dir', None))
        except (AttributeError, ImportError):
            logging.warning("{} is not a valid procedure, please double check "
                            "or register new procedure".format(procedure))

        assert result is not None, "result is mandatory for {}".format(procedure)

        def add_dependency(dag, upstream_id, downstream):
            if not dag.has_node(upstream_id):
                upstream = ProcedureNode(upstream_id, None, self)
                dag.add_node(upstream)
            else:
                upstream = dag.get_node(upstream_id)
            dag.add_dependency(upstream.node_id, downstream.node_id)

        # create inline ingredients and change definition to actual id
        for idx, ing in enumerate(ingredients):
            if isinstance(ing, dict):
                ingredient = ingredient_from_dict(dictionary=ing, **self.config)
                self.dag.add_node(IngredientNode(ingredient.id, ingredient, self))
                ingredients[idx] = ingredient.id

        if options is None:
            pdict = {'procedure': procedure, 'ingredients': ingredients, 'result': result}
        else:
            pdict = {'procedure': procedure, 'ingredients': ingredients, 'options': options, 'result': result}
        pnode = ProcedureNode(result, pdict, self)
        self.dag.add_node(pnode)
        self.cooking[collection].append(pnode.procedure)

        # adding dependencies
        for ing in ingredients:
            add_dependency(self.dag, ing, pnode)

        if options is not None:
            for ingredient_key in ['base', 'ingredient']:
                if ingredient_key in options.keys():
                    add_dependency(self.dag, options[ingredient_key], pnode)
                for opt, val in options.items():
                    if isinstance(val, Mapping) and ingredient_key in val.keys():
                        add_dependency(self.dag, options[opt][ingredient_key], pnode)
        return self

    def add_dish(self, ingredients, options=None):
        existing_dish = [x['id'] for x in self._serving]
        for ing in ingredients:
            if ing in existing_dish:
                logger.warning('dish already exist: {}, skipping...'.format(ing))
            self._serving.append({'id': ing, 'options': options})

    @staticmethod
    def register_procedure(func):
        from ddf_utils.chef import procedure as pc
        assert callable(func)
        setattr(pc, func.__name__, func)

    def run(self, serve=False, outpath=None):
        self.validate()

        results = [self.dag.get_node(x['id']).evaluate() for x in self.serving]

        if serve:
            if not outpath:
                outpath = self.config.get('out_dir', './')

            @delayed
            def _serve_all(dishes):
                for dish in dishes:
                    dish_result = self.dag.get_node(dish['id']).evaluate()

                    if 'options' in dish and dish['options'] is not None:
                        dish_result.serve(outpath, **dish['options'])
                    else:
                        dish_result.serve(outpath)

            _serve_all(self.serving).compute()

        return results

    def to_recipe(self, fp=None):
        self.validate()

        if fp is None:
            fp = sys.stdout
        recipe = dict()
        recipe['info'] = self.metadata
        recipe['config'] = self.config
        recipe['ingredients'] = list()
        recipe['cooking'] = dict()
        recipe['serving'] = self.serving

        # FIXME: ingredient should allow external csv
        for ingredient in self.ingredients:
            info = {'id': ingredient.id,
                    'dataset': ingredient.ddf_id,
                    'key': ingredient.key,
                    'values': ingredient.value,
                    }
            if ingredient.row_filter is not None:
                info['row_filter'] = ingredient.row_filter
            recipe['ingredients'].append(info)

        for k, v in self.cooking.items():
            if len(v) > 0:
                recipe['cooking'][k] = v

        yaml.round_trip_dump(recipe, fp)

    def to_graph(self, node=None):

        def process_node(g_, k_, recursive=False):
            pc = self.dag.get_node(k_).procedure['procedure']
            node_name = pc + str(time())
            g_.node(node_name, label=pc, shape='box')
            g_.edge(node_name, k_)

            for ing in self.dag.get_node(k_).get_direct_relatives(upstream=True):
                g_.edge(ing.node_id, node_name)
                if recursive:
                    process_node(g_, ing.node_id)

        g = Digraph()
        g.attr(rankdir='TB', fontsize='8')

        if node is None:
            # define all nodes
            to_serve = [x['id'] for x in self.serving]
            for k, v in self.dag.node_dict.items():
                if k in to_serve:
                    g.node(k, color='red')
                elif isinstance(self.dag.get_node(k), IngredientNode):
                    g.node(k, color='blue')
                else:
                    g.node(k)
                if isinstance(self.dag.get_node(k), ProcedureNode):
                    process_node(g, k)
        else:
            assert node in self.dag.node_dict.keys()
            g.node(node)
            process_node(g, node, recursive=True)

        return g

    @staticmethod
    def _build_recipe(recipe_file, to_disk=False, **kwargs):
        """build a complete recipe object.

        This function will check each part of recipe, convert string (the ingredient ids,
        dictionaries file names) into actual objects.

        If there are includes in recipe file, this function will run recurivly.
        If no includes found then return the parsed object as is.

        Parameters
        ----------
        recipe_file : `str`
            path to recipe file

        Keyword Args
        ------------
        to_disk : bool
            if true, save the parsed reslut to a yaml file in working dir

        Other Parameters
        ----------------
        ddf_dir : `str`
            path to search for DDF datasets, will overwrite the contfig in recipe

        """
        recipe = _loadfile(recipe_file)

        # the base dir of recipe file. for building paths for dictionary_dir and
        # sub recipe paths.
        base_dir = os.path.dirname(recipe_file)

        # the dictionary dir to retrieve translation dictionaries
        if 'config' not in recipe.keys():
            dict_dir = None
            external_csv_dir = base_dir
            recipe_dir = base_dir
        else:
            dict_dir = recipe['config'].get('dictionary_dir', None)
            external_csv_dir = recipe['config'].get('external_csv_dir', base_dir)
            recipe_dir = recipe['config'].get('recipes_dir', base_dir)

        def external_csv_abs_path(ing):
            """change the csv file in `data` to full path"""
            if isinstance(ing, dict) and 'data' in ing.keys() and isinstance(ing['data'], str):
                if not os.path.isabs(ing['data']):
                    ing['data'] = os.path.join(external_csv_dir, ing['data'])
            return ing

        # expand all files in the options
        if 'cooking' in recipe.keys():
            for p in ['concepts', 'datapoints', 'entities']:
                if p not in recipe['cooking'].keys():
                    continue
                for i, procedure in enumerate(recipe['cooking'][p]):

                    procedure['ingredients'] = [external_csv_abs_path(ing) for ing in procedure['ingredients']]
                    try:
                        opt_dict = procedure['options']['dictionary']
                    except KeyError:
                        continue
                    if isinstance(opt_dict, str):
                        # if the option dict is str, then it should be a filename
                        if dict_dir is None:
                            raise ChefRuntimeError("dictionary_dir not found in config!")
                        if os.path.isabs(dict_dir):
                            path = os.path.join(dict_dir, opt_dict)
                        else:
                            path = os.path.join(base_dir, dict_dir, opt_dict)

                        recipe['cooking'][p][i]['options']['dictionary'] = _loadfile(path)

        if 'ingredients' in recipe.keys():
            recipe['ingredients'] = [external_csv_abs_path(ing) for ing in recipe['ingredients']]

        if 'include' not in recipe.keys():
            return recipe
        else:  # append sub-recipe entities into main recipe
            sub_recipes = []
            for i in recipe['include']:
                if os.path.isabs(recipe_dir):
                    path = os.path.join(recipe_dir, i)
                else:
                    path = os.path.expanduser(os.path.join(base_dir, recipe_dir, i))
                sub_recipes.append(Chef._build_recipe(path))

            for rcp in sub_recipes:
                # appending ingredients form sub recipes
                if 'ingredients' in recipe.keys():
                    if 'ingredients' in rcp.keys():
                        [recipe['ingredients'].append(external_csv_abs_path(ing)) for ing in rcp['ingredients']]
                    # drop duplicated ingredients.
                    rcp_dict_tmp = {}
                    for v in recipe['ingredients']:
                        if v['id'] not in rcp_dict_tmp.keys():
                            rcp_dict_tmp[v['id']] = v
                        else:
                            # raise error when ingredients with same ID have different contents.
                            if v != rcp_dict_tmp[v['id']]:
                                raise ChefRuntimeError(
                                    "Different content with same ingredient id detected: " + v['id'])
                    recipe['ingredients'] = list(rcp_dict_tmp.values())
                else:
                    recipe['ingredients'] = [external_csv_abs_path(ing) for ing in rcp['ingredients']]

                # appending cooking procedures
                if 'cooking' not in rcp.keys():
                    continue
                for p in ['datapoints', 'entities', 'concepts']:
                    if p not in rcp['cooking'].keys():
                        continue
                    if 'cooking' in recipe.keys():
                        if p in recipe['cooking'].keys():
                            # NOTE: the included cooking procedures should be placed in front of
                            # the origin ones.
                            # recipe['cooking'][p] = [*rcp['cooking'][p], *recipe['cooking'][p]]
                            # ^ not supportted by Python < 3.5
                            new_procs = []
                            [new_procs.append(proc) for proc in rcp['cooking'][p]]
                            [new_procs.append(proc) for proc in recipe['cooking'][p]
                             if proc not in new_procs]
                            recipe['cooking'][p] = new_procs
                        else:
                            recipe['cooking'][p] = rcp['cooking'][p]
                    else:
                        recipe['cooking'] = {}
                        recipe['cooking'][p] = rcp['cooking'][p]

            return recipe

    @staticmethod
    def _get_dishes(recipe):
        """get all dishes in the recipe"""

        if 'serving' in recipe:
            return recipe['serving']

        dishes = list()
        for _, procs in recipe['cooking'].items():
            serve_proc_exists = False
            for p in procs:
                if p['procedure'] == 'serve':
                    serve_proc_exists = True
                    for i in p['ingredients']:
                        try:
                            dishes.append({'id': i, 'options': p['options']})
                        except KeyError:
                            dishes.append({'id': i, 'options': dict()})
            if not serve_proc_exists:
                logger.warning('no serve procedure found, will serve the last result: ' + p['result'])
                dishes.append({'id': p['result'], 'options': dict()})

        return dishes



# -*- coding: utf-8 -*-

"""Functions for scraping ILO datasets

using the bulk downloader, see `its doc`_.

.. _its doc: http://www.ilo.org/ilostat-files/WEB_bulk_download/ILOSTAT_BulkDownload_Guidelines.pdf
"""

from . common import requests_retry_session, DataFactory

from pathlib import Path
from urllib.parse import urljoin
from multiprocessing import Pool
from functools import partial

import pandas as pd


class ILOLoader(DataFactory):
    main_url = 'http://www.ilo.org/ilostat-files/WEB_bulk_download/'
    indicator_meta_url_tmpl = urljoin(main_url, 'indicator/table_of_contents_{lang}.csv')
    other_meta_url_tmpl = urljoin(main_url, 'dic/{table}_{lang}.csv')

    def load_metadata(self, table='indicator', lang='en'):
        if table == 'indicator':
            tmpl = self.indicator_meta_url_tmpl
        else:
            tmpl = self.other_meta_url_tmpl

        url = tmpl.format(table=table, lang=lang)

        metadata = {}
        metadata[table] = pd.read_csv(url)

        self.metadata = metadata[table]
        return self.metadata

    def has_newer_source(self, indicator, date):
        if self.metadata is None:
            self.load_metadata()
        md = self.metadata
        last_update = md.loc[md.id == indicator, 'last.update']
        assert len(last_update) == 1
        last_update = last_update.values[0]
        if pd.to_datetime(last_update) > pd.to_datetime(date):
            return True
        return False

    def download(self, i, out_dir):
        url = urljoin(self.main_url, f'indicator/{i}.csv.gz')
        res = requests_retry_session().get(url, stream=True, timeout=60)
        if res.status_code != 200:
            print(f'can not download source file: {url}')
            return

        with Path(out_dir, f'{i}.csv.gz').expanduser().open('wb') as f:
            for chunk in res.iter_content(chunk_size=1024):
                f.write(chunk)
                f.flush()

    def bulk_download(self, out_dir, indicators: list):

        download_ = partial(self.download, out_dir=out_dir)

        with Pool(5) as p:
            p.map(download_, indicators)



# -*- coding: utf-8 -*-

from . cdiac import CDIACLoader
from . clio_infra import ClioInfraLoader
from . igme import IGMELoader
from . ihme import IHMELoader
from . ilo import ILOLoader
from . oecd import OECDLoader
from . worldbank import WorldBankLoader


# -*- coding: utf-8 -*-

from functools import wraps, partial
import json
from .. import config
from .. import ops
import os
import logging
import click
import numpy as np


def prompt_select(selects, text_before=None):
    """ask user to choose in a list of options"""
    def value_proc(v):
        if v == 'n':
            return -1
        if v == 'q':
            import sys
            sys.exit(130)  # code for user interrupted
        return int(v)

    if text_before:
        click.echo(text_before)
    select_text = []

    for i, v in enumerate(selects):
        select_text.append('{}: {}'.format(i+1, v))
    select_text = '\n'.join(select_text)
    click.echo(select_text)

    prompt_text = 'Please select a value ({} ~ {}), or "n" to skip, "q" to quit'\
                  .format(1, len(selects))
    val = click.prompt(prompt_text, value_proc=value_proc)
    if val == -1:
        return -1
    return selects[val-1]


def read_opt(options, key, required=False, default=None):
    """utility to read an attribute from an options dictionary

    Parameters
    ----------
    options : dict
        the option dictionary to read
    key : `str`
        the key to read

    Keyword Args
    ------------
    required : bool
        if true, raise error if the `key` is not in the option dict
    default : object
        a default to return if `key` is not in option dict and `required` is false
    """
    if key in options.keys():
        return options.pop(key)
    if required:
        raise KeyError('field {} is mandantory'.format(key))
    return default


def accept_file_input(params):
    """read json file for parameters for the function"""
    def real_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for k, v in kwargs.items():
                if k in params:
                    if isinstance(v, str):
                        assert os.path.exists(v), 'file not exists!'
                        kwargs[k] = json.load(open(v))
            return func(*args, **kwargs)
        return wrapper
    return real_wrapper


def mkfunc(options):
    """create function warppers base on the options provided

    This function is used in procedures which have a function block. Such as
    :py:func:`ddf_utils.chef.procedure.groupby`. It will try to return functions
    from numpy or :py:mod:`ddf_utils.ops`.

    Parameters
    ----------
    options : `str` or dict
        if a dictionary provided, "function" should be a key in the dictionary
    """
    if isinstance(options, str):
        return getattr(np, options)
    else:
        func = getattr(ops, options.pop('function'))
        return partial(func, **options)


# below functions are not used in ddf_utils yet, but may be useful.
def log_shape(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        logging.info("%s,%s" % (func.__name__, result.shape))
        return result
    return wrapper


def log_dtypes(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        logging.info("%s,%s" % (func.__name__, result.dtypes))
        return result
    return wrapper


def log_procedure(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        logging.info("running %s" % (func.__name__))
        result = func(*args, **kwargs)
        return result
    return wrapper


def debuggable(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if 'debug' in kwargs.keys():
            debug = kwargs.pop('debug')
            result = func(*args, **kwargs)
            if debug:
                if config.DEBUG_OUTPUT_PATH is None:
                    logging.warning('debug output path not set!')
                    config.DEBUG_OUTPUT_PATH = './_debug'  # TODO: handle the config better
                outpath = os.path.join(config.DEBUG_OUTPUT_PATH, result.ingred_id)
                if os.path.exists(outpath):
                    import shutil
                    shutil.rmtree(outpath)
                    os.mkdir(outpath)
                result.serve(outpath)
        else:
            result = func(*args, **kwargs)
        return result
    return wrapper



# -*- coding: utf-8 -*-

"""Functions to load data from Worldbank API.

We use its bulkdownload utilities.

Source link: `WorldBank website`_

.. _WorldBank website: http://data.worldbank.org

"""

import os.path as osp

import requests
import pandas as pd

from urllib.parse import urlsplit
from concurrent.futures import ThreadPoolExecutor, as_completed

from . common import DataFactory


class WorldBankLoader(DataFactory):
    __doc__ = """T.B.D"""
    url = 'http://api.worldbank.org/v2/datacatalog?format=json'

    def load_metadata(self):
        def load_url(url, page=None):
            if page:
                url = url + '&page=' + str(page)
            res = requests.get(url)
            return res.json()

        def handle_data(data):
            for v in data['datacatalog']:
                d = {}
                for v_ in v['metatype']:
                    d[v_['id']] = v_['value']
                yield d

        first_page = self.url
        data = load_url(first_page)

        metadata = []

        for d in handle_data(data):
            metadata.append(d)

        pages = range(2, data['pages']+1)

        executor = ThreadPoolExecutor(max_workers=3)
        res = [executor.submit(load_url, self.url, page=p) for p in pages]

        for r in as_completed(res):
            for d in handle_data(r.result()):
                # print(d)
                metadata.append(d)

        metadata = pd.DataFrame.from_records(metadata)
        self.metadata = metadata
        return metadata

    def has_newer_source(self, dataset, date):
        if self.metadata is None:
            self.load_metadata()
        metadata = self.metadata
        lastrevisiondate = metadata.loc[
            metadata.acronym == dataset,
            'lastrevisiondate'].values[0]
        if pd.to_datetime(lastrevisiondate) > pd.to_datetime(date):
            return True
        else:
            return False

    def bulk_download(self, dataset, out_dir):
        if self.metadata is None:
            self.load_metadata()
        metadata = self.metadata
        s = metadata.loc[metadata.acronym == dataset, 'bulkdownload'].values[0]
        url = s.split(';')[1].split('=')[1]
        res = requests.get(url, stream=True)

        fs_path = osp.join(out_dir, osp.basename(urlsplit(url).path))

        with open(fs_path, 'wb') as f:
            for c in res.iter_content(chunk_size=1024):
                f.write(c)
            f.close()

        return out_dir



"""APIs for chef"""

import coloredlogs

from . model.chef import Chef
from . model.ingredient import ingredient_from_dict


coloredlogs.install(level=None, fmt='%(asctime)s %(levelname)s %(message)s')


def run_recipe(fn, ddf_dir, out_dir):
    """run the recipe file and serve result"""
    from ddf_utils.io import cleanup
    import os
    if os.path.exists(out_dir):
        cleanup(out_dir)
    else:
        os.mkdir(out_dir)

    chef = Chef.from_recipe(fn, ddf_dir=ddf_dir)
    chef.run(serve=True, outpath=out_dir)
    return



# -*- coding: utf-8 -*-

"""QA functioins.
"""

import sys
import logging
import pandas as pd
import numpy as np

logger = logging.getLogger('QA')
this = sys.modules[__name__]


def _gen_indicator_key_list(ddf):
    for indicator, key_data_pair in ddf.datapoints.items():
        for key in key_data_pair.keys():
            yield(indicator, key)


def compare_with_func(dataset1, dataset2, fns=None,
                      indicators=None, key=None, **kwargs):
    """compare 2 datasets with functions"""

    if not fns:
        fns = ['rmse', 'nrmse']

    indicators1 = list(_gen_indicator_key_list(dataset1))
    indicators2 = list(_gen_indicator_key_list(dataset2))

    # check availability for indicators
    s1 = set(indicators1)
    s2 = set(indicators2)

    diff12 = s1 - s2
    diff21 = s2 - s1

    if len(diff12) > 0:
        msg = ["below indicators are only available in {}".format(dataset1.props['name'])]
        for item in diff12:
            msg.append("- {} by {}".format(item[0], ', '.join(item[1])))
        msg.append('')
        logger.warning('\n'.join(msg))
    if len(diff21) > 0:
        msg = ["below indicators are only available in {}".format(dataset2.props['name'])]
        for item in diff21:
            msg.append("- {} by {}".format(item[0], ', '.join(item[1])))
        msg.append('')
        logger.warning('\n'.join(msg))

    # construct a dataframe, including all indicators in both dataset.
    result = pd.DataFrame(list(s1.union(s2)), columns=['indicator', 'primary_key'])

    def get_comp_df(indicator, k):
        """get dataframes from old and new datasets, and combine them into one dataframe"""
        # FIXME: support multiple indicator in one file
        # like the indicators in ddf--sodertorn--stockholm_lan_basomrade
        try:
            i1 = dataset1.get_datapoints(indicator, k).data.compute().set_index(list(k))
        except KeyError:
            raise
        try:
            i2 = dataset2.get_datapoints(indicator, k).data.compute().set_index(list(k))
        except KeyError:
            raise
        # i1 = i1.rename(columns={indicator: 'old'})
        # i2 = i2.rename(columns={indicator: 'new'})
        # comp = pd.concat([i1, i2], axis=1)
        comp = i1.join(i2, how='outer', lsuffix='_old', rsuffix='_new')

        return comp

    def do_compare(fns, indicator, k):
        try:
            comp_df = get_comp_df(indicator, k)
        except KeyError:
            return [np.nan] * len(fns)

        # return nan if indicator is not number type
        if comp_df.dtypes[indicator+'_old'] == 'object' or comp_df.dtypes[indicator+'_new'] == 'object':
            return [np.nan] * len(fns)

        return [f(comp_df, indicator, **kwargs) if callable(f) else getattr(this, f)(comp_df, indicator, **kwargs)
                for f in fns]

    # only keep indicators we want to compare
    if indicators:
        result = result[result.indicator.isin(indicators)]
    if key:
        result = result[result.primary_key.isin(key)]

    # append new columns before we do calculation
    for f in fns:
        result[f] = np.nan

    result = result.set_index(['indicator', 'primary_key'])
    result = result.sort_index()
    idx = pd.IndexSlice
    for i in result.index:
        result.loc[idx[i[0], [i[1]]], fns] = np.array(do_compare(fns, i[0], i[1]))

    return result.reset_index()


def rval(comp_df, indicator, on='geo'):
    """return r-value between old and new data"""
    # FIXME: how to set the on parameter from cli?
    old_name = indicator+'_old'
    new_name = indicator+'_new'
    # logger.warning("{}".format(old_name, new_name))
    # logger.warning("{}".format(comp_df.columns))

    def f(df):
        df_ = df - df.shift(1)
        r = df_.corr().loc[old_name, new_name]
        return r

    level = comp_df.index.names.index(on)
    res = comp_df.groupby(level=level).apply(f)

    return res.mean()


def avg_pct_chg(comp_df, indicator, on='geo'):
    """return average precentage changes between old and new data"""
    # FIXME: how to set the on parameter from cli?
    old_name = indicator+'_old'
    new_name = indicator+'_new'
    level = comp_df.index.names.index(on)

    def f(df):
        new = df[new_name]
        old = df[old_name]
        chg = (new - old) / old * 100

        return chg.replace([np.inf, -np.inf], np.nan).mean()

    res = comp_df.groupby(level=level).apply(f)

    return res.abs().mean()


def max_pct_chg(comp_df, indicator, **kwargs):
    """return average precentage changes between old and new data"""
    old_name = indicator+'_old'
    new_name = indicator+'_new'
    res = (comp_df[new_name] - comp_df[old_name]) / comp_df[old_name] * 100
    res = res.replace([np.inf, -np.inf], np.nan)
    return res.abs().max()


# def min_pct_chg(comp_df, indicator):
#     """return average precentage changes between old and new data"""
#     old_name = indicator+'_old'
#     new_name = indicator+'_new'
#     res = (comp_df[new_name] - comp_df[old_name]) / comp_df[old_name] * 100
#     res = res.replace([np.inf, -np.inf], np.nan)
#     return res.min()


def max_change_index(comp_df, indicator, **kwargs):
    # FIXME: this function makes all result column type to be object
    # see test cases in test_qa.py
    old_name = indicator+'_old'
    new_name = indicator+'_new'
    diff = (comp_df[new_name] - comp_df[old_name]) / comp_df[old_name] * 100
    diff = diff.replace([np.inf, -np.inf], np.nan)
    diff = abs(diff)
    if len(diff.dropna()) == 0 or diff.max() == 0:
        return ''
    idx = diff[diff == diff.max()].index.values[0]
    return str(idx)


def rmse(comp_df, indicator, **kwargs):
    old_name = indicator+'_old'
    new_name = indicator+'_new'
    diff_2 = np.power(comp_df[new_name] - comp_df[old_name], 2)
    rmse_val = np.sqrt(np.sum(diff_2) / len(diff_2))

    return rmse_val


def nrmse(comp_df, indicator, **kwargs):
    old_name = indicator+'_old'
    new_name = indicator+'_new'

    rmse_val = rmse(comp_df, indicator)
    nrmse_val = rmse_val / (comp_df[old_name].max() - comp_df[new_name].min())

    return nrmse_val


def new_datapoints(comp_df, indicator, **kwargs):
    old_name = indicator+'_old'
    new_name = indicator+'_new'

    count = comp_df[new_name].isnull().sum()
    return count


def dropped_datapoints(comp_df, indicator, **kwargs):
    old_name = indicator+'_old'
    new_name = indicator+'_new'

    count = comp_df[old_name].isnull().sum()
    return count



# -*- coding: utf-8 -*-

"""Functions for scraping OECD website using their SDMX API

source link `OECD website`_

.. _OECD website: http://stats.oecd.org

"""

from io import BytesIO
from pathlib import Path
import json

import requests
from lxml import etree
import pandas as pd

from .common import DataFactory


class OECDLoader(DataFactory):
    # TODO: there is 1 000 000 row limitation. handle it
    data_url_tmpl = 'http://stats.oecd.org/SDMX-JSON/data/{dataset}/all/all'
    # dataflow_url_tmpl = 'http://stats.oecd.org/SDMX-JSON/dataflow/{dataset}/all/all'
    datastructure_url_tmpl = 'http://stats.oecd.org/restsdmx/sdmx.ashx/GetDataStructure/{dataset}'
    metadata_url = 'http://stats.oecd.org/RestSDMX/sdmx.ashx/GetKeyFamily/all'

    def load_metadata(self):
        res = requests.get(self.metadata_url)
        parser = etree.XMLParser(ns_clean=True)  # TODO: ns_clean can't clean namespace. Why?
        tree = etree.parse(BytesIO(res.content), parser)
        root = tree.getroot()

        datasets = []

        for e in root.findall('.//KeyFamily', namespaces=root.nsmap):
            dataset_id = (e.attrib['id'])
            dataset_name = None
            for e_ in e.findall('.//Name', namespaces=root.nsmap):
                if 'en' in e_.attrib.values():
                    dataset_name = e_.text

            datasets.append({'id': dataset_id, 'name': dataset_name})

        metadata = pd.DataFrame.from_records(datasets)
        self.metadata = metadata
        return metadata

    def has_newer_source(self, dataset, version):
        # There is no version info in the response from OECD server.
        return NotImplementedError

    def bulk_download(self, out_dir, dataset):
        """download the full json, including observation/dimension lists."""
        p = Path(out_dir, dataset+'.json')
        url = self.data_url_tmpl.format(dataset=dataset)

        res = requests.get(url)
        assert res.status_code == 200, 'error: {}'.format(res.status_code)

        with p.expanduser().open('wb') as f:
            f.write(res.content)
            f.close()



# -*- coding: utf-8 -*-

"""The DDF model"""

import os.path as osp
from typing import List, Tuple, Dict, Union, Callable
from abc import ABC, abstractmethod
import attr
import json
from pathlib import Path
from itertools import product
from tqdm import tqdm
from collections import OrderedDict

import numpy as np
import pandas as pd
import dask.dataframe as dd

import logging

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True)
class Concept:
    id: str
    concept_type: str
    props: dict = attr.ib(factory=dict)

    def to_dict(self):
        res = OrderedDict()
        res['concept'] = self.id
        res['concept_type'] = self.concept_type
        props = self.props.copy()
        for k, v in props.items():
            res[k] = v
        return res


@attr.s(auto_attribs=True)
class Entity:
    id: str
    domain: str
    sets: List[str]
    props: dict = attr.ib(factory=dict)

    def to_dict(self, pkey=None):
        """create a dictionary containing name/domain/is--headers/and properties
        So this can be easily plug in pandas.DataFrame.from_records()
        """
        res = OrderedDict()
        if pkey:
            res[pkey] = self.id
        else:
            res[self.domain] = self.id
        if self.sets:
            for s in self.sets:
                header = f'is--{s}'
                res[header] = 'TRUE'
        props = self.props.copy()
        for k, v in props.items():
            res[k] = v

        return res


# @attr.s(auto_attribs=True)
# class EntitySet:
#     id: str
#     entities: List[Entity]
#     domain: str
#     props: dict = attr.ib(factory=dict)


@attr.s(auto_attribs=True)
class EntityDomain:
    id: str
    entities: List[Entity]
    props: dict = attr.ib(factory=dict)

    @property
    def entity_sets(self):
        sets = set()
        for e in self.entities:
            for s in e.sets:
                sets.add(s)
        return list(sets)

    def get_entity_set(self, s):
        return [e for e in self.entities if s in e.sets]

    def to_dict(self, eset=None):
        if eset:
            entities = self.get_entity_set(eset)
            return [e.to_dict(eset) for e in entities]
        else:
            defaultdict = dict()
            for s in self.entity_sets:
                header = f'is--{s}'
                defaultdict[header] = 'FALSE'
            entities = self.entities
            res = list()
            for e in entities:
                d = e.to_dict()
                # appending False into the is--headers
                for k, v in defaultdict.items():
                    d.setdefault(k, v)
                res.append(d)
            return res

    def add_entity(self, ent: Entity):
        if ent.domain != self.id:
            raise ValueError('domain name mismatch for the entity object and domain object!')
        for existing_ent in self.entities:
            if ent.id == existing_ent.id:
                logger.debug('updating existing entity: {}'.format(existing_ent.id))
                for s in ent.sets:
                    if s not in existing_ent.sets:
                        existing_ent.sets.append(s)
                # TODO: logging for existing fileds
                existing_ent.props.update(ent.props)
                break
        else:
            logger.debug('appending entity: {}'.format(ent.id))
            self.entities.append(ent)

    def __getitem__(self, item):
        return self.get_entity_set(item)


@attr.s(auto_attribs=True)
class DataPoint(ABC):
    """A DataPoint object stores a set of datapoints which have same dimensions and
    which belongs to only one indicator."""
    id: str
    dimensions: Tuple[str]
    store: str
    # TODO: think about cache

    @property
    @abstractmethod
    def data(self):
        ...


@attr.s
class PandasDataPoint(DataPoint):
    """load datapoints with pandas"""
    path: str = attr.ib()
    dtypes: dict = attr.ib()
    store = attr.ib(default='pandas')
    # data_cache = attr.ib()

    @property
    def data(self):
        cols = [*self.dimensions, self.id]
        return pd.read_csv(self.path, dtype=self.dtypes)[cols]


@attr.s
class DaskDataPoint(DataPoint):
    """load datapoints with dask"""
    path: Union[List[str], str] = attr.ib()  # can be a list of paths
    dtypes: dict = attr.ib()
    store = attr.ib(default='dask')

    @property
    def data(self):
        cols = [*self.dimensions, self.id]
        return dd.read_csv(self.path, dtype=self.dtypes)[cols]


@attr.s(auto_attribs=True, repr=False)
class Synonym:
    concept_id: str
    synonyms: Dict[str, str]

    def __repr__(self):
        dict_str = self.synonyms.__str__()
        if len(dict_str) > 20:
            dict_str = dict_str[:20] + ' ... }'
        return "Synonym(concept_id={}, synonyms={})".format(self.concept_id, dict_str)

    def to_dict(self):
        pass


@attr.s(auto_attribs=True, repr=False)
class DDF:
    # Here I use dictionaries for the data structure, just for performance
    # in fact they can be just lists, i.e. concepts is just a list of Concept objects.
    concepts: Dict[str, Concept]
    entities: Dict[str, EntityDomain] = attr.ib(factory=dict)
    datapoints: Dict[str, Dict[str, DataPoint]] = attr.ib(factory=dict)
    synonyms: Dict[str, Synonym] = attr.ib(factory=dict)
    props: dict = attr.ib(factory=dict)

    def __repr__(self):
        # def maybe_truncate(obj, maxlen=20, fillspaces=False):
        #     if isinstance(obj, np.ndarray):
        #         s = ', '.join(map(str, obj))
        #         if len(s) > maxlen:
        #             s = ','.join(s[:(maxlen - 3)].split(',')[:-1])
        #             s = s + '...'
        #     else:
        #         s = str(obj)
        #         if len(s) > maxlen:
        #             s = s[:(maxlen - 3)] + '...'
        #     if len(s) < maxlen and fillspaces:
        #         diff = maxlen - len(s)
        #         s = s + ' ' * diff
        #     return s

        indent = 4

        # TODO: also report domains/datapoints defined in concepts but without data.
        docs = ["<Dataset {}>".format(self.props.get('name', 'NONAME'))]
        docs.append("entity domains:")
        for domain_id, domain in self.entities.items():
            if not domain.entity_sets:
                docs.append(f'{" " * indent}* {domain_id}')
            else:
                docs.append(f'{" " * indent}* {domain_id}:')
                for ent_set in domain.entity_sets:
                    docs.append(f'{" " * indent * 2}- {ent_set}')
        docs.append("datapoints:")
        for indicator_id, ind_dict in self.datapoints.items():
            for pkey, _ in ind_dict.items():
                pkey_str = ', '.join(pkey)
                docs.append(f'{" " * indent}- ({pkey_str}) {indicator_id}')

        return '\n'.join(docs)

    def indicators(self, by=None):
        # TODO: naming
        if not by:
            return list(self.datapoints.keys())
        else:
            res = list()
            for i, v in self.datapoints.items():
                if tuple(sorted(by)) in v.keys():
                    res.append(i)
            return res

    def get_datapoints(self, i, by=None):
        # TODO: query by entity domain should work too.
        # current only entity set can be queried.
        if by:
            by_ = tuple(sorted(by))
        else:
            if len(self.datapoints[i]) > 1:
                raise ValueError("there are multiple primary keys for this indicator, "
                                 "you should provide the primary key")
            by_ = list(self.datapoints[i].keys())[0]

        return self.datapoints[i][by_]

    def get_entities(self, domain, eset=None):
        if not eset:
            return self.entities[domain].entities
        else:
            return [e for e in self.entities[domain].entities if eset in e.sets]

    # TODO: maybe add below methods to modify DDF objects.
    # def get_concept_ids(self, concept_type=None):
    #     if concept_type is not None:
    #         return [concept.id for concept in self.concepts if concept.concept_type == concept_type]
    #     return [concept.id for concept in self.concepts]
    #
    # def get_indicators(self):
    #     dps = [(dp.id, dp.dimensions) for dp in self.datapoints]
    #
    # def add_concept(self, c: Concept):
    #     self.concepts.append(c)
    #
    # def remove_concept(self, c_id: str):
    #     self.concepts = [concept for concept in self.concepts if concept.id != c_id]
    #
    # def add_entity_domain(self, e: EntityDomain):
    #     self.entities.append(e)
    #
    # def remove_entity_domain(self, e_id: str):
    #     pass
    #
    # def add_datapoints(self, dps: DataPoint):
    #     self.datapoints.append(dps)
    #
    # def remove_datapoints(self, dps_id: str):
    #     pass

    # def validate(self):  # should be a function outside the class.
    #     """check if the DDF object is valid.
    #
    #     1. datapoints's dimensions should be time or entity domains/sets
    #     2. datapoints concepts and entity concepts should exists in concepts
    #     3. call every object's validate method.
    #     """
    #     pass
    #
    # def create_concepts(self):
    #     """create concepts from existing entity_domain/entity_set/datapoints"""
    #     pass



# -*- coding: utf-8 -*-

"""functions for handling DDF datapackage"""

import os
import os.path as osp
import json
import csv
import re
from datetime import datetime, timezone

from collections.abc import Mapping, Sequence
from collections import OrderedDict
from itertools import product

import dask.dataframe as dd
import pandas as pd

from tqdm import tqdm

from .model.package import DDFcsv
from .model.utils import sort_json
from ddf_utils.chef.helpers import read_opt

import logging


logger = logging.getLogger(__name__)


# check if a directory is dataset root dir
def is_datapackage(path):
    """check if a directory is a dataset directory

    This function checks if ddf--index.csv and datapackage.json exists
    to judge if the dir is a dataset.
    """
    index_path = osp.join(path, 'ddf--index.csv')
    datapackage_path = osp.join(path, 'datapackage.json')
    if osp.exists(index_path) or osp.exists(datapackage_path):
        return True
    else:
        return False


def get_datapackage(path, use_existing=True, update=False):
    """get the datapackage.json from a dataset path, create one if it's not exists

    Parameters
    ----------
    path : `str`
        the dataset path

    Keyword Args
    ------------
    use_existing : bool
        whether or not to use the existing datapackage
    update : bool
        if update is true, will update the resources and schema in existing datapackage.json. else just return existing
        datapackage.json
    """
    datapackage_path = os.path.join(path, 'datapackage.json')

    if os.path.exists(datapackage_path):
        with open(datapackage_path, encoding='utf8') as f:
            datapackage_old = json.load(f, object_pairs_hook=OrderedDict)

        if use_existing:
            if not update:
                return datapackage_old
            try:
                datapackage_old.pop('resources')  # don't use the old resources
                datapackage_old.pop('ddfSchema')  # and ddf schema
            except KeyError:
                logger.warning('no resources or ddfSchema in datapackage.json')
            datapackage_new = create_datapackage(path, **datapackage_old)
        else:
            datapackage_new = create_datapackage(path)
    else:
        if use_existing:
            print("WARNING: no existing datapackage.json")
        datapackage_new = create_datapackage(path)

    return datapackage_new


def create_datapackage(path, gen_schema=True, **kwargs):
    """create datapackage.json base on the files in `path`.

    If you want to set some attributes manually, you can pass them as
    keyword arguments to this function

    Note
    ----
    A DDFcsv datapackage MUST contain the fields `name` and `resources`.

    if name is not provided, then the base name of `path` will be used.

    Parameters
    ----------
    path : `str`
        the dataset path to create datapackage.json
    gen_schema : bool
        whether to create DDFSchema in datapackage.json. Default is True
    kwargs : dict
        metadata to write into datapackage.json. According to spec,
        title, description, author and license SHOULD be fields in datapackage.json.
    """

    datapackage = OrderedDict()

    # setting default fields
    default_name = os.path.basename(os.path.normpath(os.path.abspath(path)))
    name = read_opt(kwargs, 'name', default=default_name, method='pop')
    title = read_opt(kwargs, 'title', default=default_name, method='pop')
    description = read_opt(kwargs, 'description', default='', method='pop')
    author = read_opt(kwargs, 'author', default='', method='pop')
    dp_license = read_opt(kwargs, 'license', default='', method='pop')
    default_lang = {'id': 'en'}
    lang = read_opt(kwargs, 'lang', default=default_lang, method='pop')

    datapackage['name'] = name
    datapackage['language'] = lang
    datapackage['title'] = title
    datapackage['description'] = description
    datapackage['author'] = author
    datapackage['license'] = dp_license

    # add all optional settings
    for k in sorted(kwargs.keys()):
        datapackage[k] = kwargs[k]

    # update the last updated time
    datapackage['created'] = datetime.utcnow().replace(tzinfo=timezone.utc).isoformat()

    # generate resources
    resources = []
    names_sofar = dict()

    for f in get_ddf_files(path):
        path_res = f
        name_res = os.path.splitext(os.path.basename(f))[0]

        if name_res in names_sofar.keys():
            names_sofar[name_res] = names_sofar[name_res] + 1
            # adding a tail to the resource name, because it should be unique
            name_res = name_res + '-' + str(names_sofar[name_res])
        else:
            names_sofar[name_res] = 0

        resources.append(OrderedDict([('path', path_res), ('name', name_res)]))

    for n, r in enumerate(resources):
        name_res = r['name']
        path_res = r['path']
        schema = {"fields": [], "primaryKey": None}

        if 'datapoints' in name_res:
            keys = re.match(r'ddf--datapoints--.*--by--(.*)', name_res).groups()[0]
            primary_keys = keys.split('--')
            # print(conc, primary_keys)
            for i, k in enumerate(primary_keys):
                if '-' in k:
                    k_new, *enums = k.split('-')
                    primary_keys[i] = k_new
                    constraint = {'enum': enums}
                    schema['fields'].append({'name': k_new, 'constraints': constraint})
                else:
                    schema['fields'].append({'name': k})

            with open(os.path.join(path, path_res)) as f:
                headers_line = f.readline()
                f.close()

            headers_line = headers_line.strip('\n')
            headers = headers_line.split(',')
            headers = [x.strip() for x in headers]
            headers = set(headers)
            fields = headers.difference(set(primary_keys))

            for field in fields:
                schema['fields'].append({'name': field})
            schema['primaryKey'] = primary_keys

            resources[n].update({'schema': schema})

        elif 'entities' in name_res:
            match = re.match(r'ddf--entities--([\w_]+)(--[\w_]*)?-?.*', name_res).groups()
            domain, concept = match
            if concept is not None:
                concept = concept[2:]

            with open(os.path.join(path, r['path'])) as f:
                reader = csv.reader(f, delimiter=',', quotechar='"')
                # we only need the headers for index file
                header = next(reader)

            if domain in header:
                key = domain
            elif concept is not None and concept in header:
                key = concept
            else:
                raise ValueError('no header in {} matches its implied domain/entity_set!'.format(name_res))
                # print(
                #     """There is no matching header found for {}. Using the first column header
                #     """.format(name_res)
                # )
                # key = header[0]

            schema['primaryKey'] = key
            for h in header:
                schema['fields'].append({'name': h})
            resources[n].update({'schema': schema})

        elif 'concepts' in name_res:
            with open(os.path.join(path, r['path'])) as f:
                reader = csv.reader(f, delimiter=',', quotechar='"')
                header = next(reader)
            schema['primaryKey'] = 'concept'
            for h in header:
                schema['fields'].append({'name': h})

            resources[n].update({'schema': schema})
        elif 'synonyms' in name_res:
            with open(os.path.join(path, r['path'])) as f:
                reader = csv.reader(f, delimiter=',', quotechar='"')
                header = next(reader)
            k1 = 'synonym'
            k2 = r['name'].split('--')[-1]
            schema['primaryKey'] = [k1, k2]
            for h in header:
                schema['fields'].append({'name': h})

            resources[n].update({'schema': schema})
        else:  # not entity/concept/datapoint. it's not supported yet so we don't include them.
            print("not supported file: " + name_res)
            resources[n] = None

    datapackage['resources'] = [x for x in resources if x is not None]

    # generate ddf schema
    if gen_schema:
        dp = DDFcsv.from_dict(datapackage, base_path=path)
        logger.info('generating ddf schema, may take some time...')
        dp.ddfSchema = dp.generate_ddf_schema()
        result = dp.to_dict()
    else:
        result = datapackage

    return sort_json(result)


def get_ddf_files(path, root=None):
    """yield all csv files which are named following the DDF model standard.

    Parameters
    -----------
    path : `str`
        the path to check
    root : `str`, optional
        if path is relative, append the root to all files.
    """
    info = next(os.walk(path))

    # don't include hidden and lang/etl dir.
    sub_dirs = [
        x for x in info[1] if (not x.startswith('.') and x not in ['lang', 'etl', 'langsplit'])
    ]
    files = list()
    for x in info[2]:
        if x.startswith('ddf--') and x != 'ddf--index.csv' and x.endswith('.csv'):
            files.append(x)
        else:
            logging.warning('skipping file {}'.format(x))

    for f in files:
        if root:
            yield os.path.join(root, f)
        else:
            yield f

    for sd in sub_dirs:
        for p in get_ddf_files(os.path.join(path, sd), root=sd):
            yield p



# -*- coding: utf-8 -*-

"""global variables for ddf_utils"""

import os

DDF_SEARCH_PATH = os.getenv('DDF_SEARCH_PATH')

if not DDF_SEARCH_PATH:
    DDF_SEARCH_PATH = './'

DICT_PATH = None
DEBUG_ALL = False
DEBUG_OUTPUT_PATH = None



# -*- coding: utf-8 -*-

"""procedures, base on dataset"""

import logging
from .helpers import accept_file_input
from ..model.ddf import Dataset

logging.getLogger('Chef')


@accept_file_input(['dictionary'])
def translate_concept(chef, dataset=None, dictionary=None) -> Dataset:
    """translate concepts, base on dictionary"""
    ds = chef.get_ingredient(dataset)
    return ds.rename(concepts=dictionary)


def _gen_mapping_dict(df, column, dictionary):
    pass


def translate_entity(dataset: Dataset, dictionary, column, *,
                     target_column=None, not_found='drop',
                     ambiguity='prompt', ignore_case=False):
    """translate entity values base on dictionary"""
    if sorted(dictionary.keys()) == sorted(['base', 'key', 'value']):
        dictionary = _gen_mapping_dict(dataset.get_entity(target_column), target_column, dictionary)

    raise NotImplementedError


def split_entity(dataset, dictionary, target_entity, splitted='drop'):
    """split entity to new entities"""

    from ..transformer import split_keys

    concepts, entities, datapoints = dataset.get_data_copy()
    new_entities = set()

    # update datapoints
    for indicator, kvs in datapoints.items():
        for k, df in kvs.items():
            if target_entity in k:
                kvs[k] = split_keys(df.set_index(k),
                                    target_entity, dictionary, splitted).reset_index()
                new_entities = new_entities.union(set(kvs[k][target_entity].unique()))

    # update entities
    if target_entity in dataset.domains:
        pass


def merge_entity(dataset):
    pass


def trend_bridge(dataset):
    pass


def run_op(dataset):
    pass


def window(dataset):
    pass


def groupby(dataset):
    pass


def filter_row(dataset):
    pass


def merge(dataset):
    pass


# functions for add datapoints/concepts/entities
#


# -*- coding: utf-8 -*-

"""translate_header procedures for recipes"""

import logging
from typing import List

from .. helpers import debuggable, build_dictionary
from .. model.ingredient import Ingredient, get_ingredient_class
from .. exceptions import ProcedureError
from .. model.chef import Chef


logger = logging.getLogger('translate_header')


def _translate_header_datapoints(ingredient, dictionary, duplicated):
    """Translating Datapoints.

    1. rename the columns in dataframe
    2. check duplication in the origin data.keys()
    3. rename the key in data.keys()
    """
    data = ingredient.get_data()
    new_data = dict()

    for k in list(data.keys()):
        df_new = data[k].rename(columns=dictionary)

        if k not in dictionary.keys():  # no need to translate keys in `data`
            new_data[k] = df_new
            continue

        new_name = dictionary[k]
        if new_name in data.keys():  # duplicated
            if duplicated == 'error':
                raise ValueError(
                    'can not rename indicator {} to {} '
                    'because {} already exists!'.format(k, new_name, new_name))
            else:
                # replace
                new_data[new_name] = df_new
        else:
            new_data[new_name] = df_new

    return new_data


def _translate_header_entities_concepts(ingredient, dictionary, duplicated):
    """Translating entities or concepts

    Entities and concepts only have one dataframe inside the ingredient, so renaming will
    happen in the dataframe columns.

    Note that it's not allowed to translate reversed keyword `concept`/`concept_type`
    """
    data = ingredient.get_data()
    new_data = dict()

    if 'concept' in dictionary and ingredient.dtype == 'concepts':
        raise ValueError('cannot translate the primaryKey for concept!')

    for k in list(data.keys()):
        df_new = data[k].copy()
        for old_name, new_name in dictionary.items():
            if new_name in df_new.columns:
                if duplicated == 'error':
                    raise ValueError(
                        'can not rename column {} to {} '
                        'because {} already exists!'.format(old_name, new_name, new_name))
                elif duplicated == 'replace':
                    df_new[new_name] = df_new[old_name]
                    df_new = df_new.drop(old_name, axis=1)
                else:
                    raise ValueError('unknown option to `duplicated`: {}'.format(duplicated))
            elif old_name in df_new.columns:
                df_new[new_name] = df_new[old_name]
                df_new = df_new.drop(old_name, axis=1)
            else:
                continue

        if ingredient.dtype == 'entities':  # also rename the `is--` headers
            rm_ = {}
            for c in df_new.columns:
                if c.startswith('is--'):
                    old_name_is_header = c[4:]
                    if old_name_is_header in dictionary:
                        rm_[c] = 'is--' + dictionary[old_name_is_header]
            if len(rm_) > 0:
                df_new = df_new.rename(columns=rm_)
        if k in dictionary.keys():  # if we need to rename the concept name
            new_data[dictionary[k]] = df_new
            # del(data[k])
        else:  # we only rename index/properties columns
            new_data[k] = df_new

    return new_data


@debuggable
def translate_header(chef: Chef, ingredients: List[Ingredient],
                     result, dictionary, duplicated='error') -> Ingredient:
    """Translate column headers

    Procedure format:

    .. code-block:: yaml

       procedure: translate_header
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingredient id
       options:
         dictionary: str or dict  # file name or mappings dictionary

    Parameters
    ----------
    chef : Chef
        The Chef the procedure will run on
    ingredients : list
        A list of ingredient id in the dag to translate
    dictionary : dict or `str`
        A dictionary for name mapping, or filepath to the dictionary
    duplicated : `str`
       What to do when there are duplicated columns after renaming. Avaliable options
       are `error`, `replace`
    result : `str`
        The result ingredient id

    See Also
    --------
    :py:func:`ddf_utils.transformer.translate_header` : Related function in transformer module
    """
    assert len(ingredients) == 1, "procedure only support 1 ingredient for now."
    ingredient = ingredients[0]
    logger.info("translate_header: " + ingredient.id)

    d = build_dictionary(chef, dictionary)

    if ingredient.dtype == 'datapoints':
        new_data = _translate_header_datapoints(ingredient, d, duplicated)
    elif ingredient.dtype in ['entities', 'concepts']:
        new_data = _translate_header_entities_concepts(ingredient, d, duplicated)
    else:
        raise ProcedureError("translate header do not support this data type: {}".format(ingredient.dtype))

    # also rename the key
    if ingredient.dtype == 'datapoints':
        newkey = ingredient.key.copy()
        for key in d.keys():
            if key in ingredient.key:
                newkey[newkey.index(key)] = d[key]
    elif ingredient.dtype == 'entities':
        newkey = ingredient.key
        for key in d.keys():
            if key == ingredient.key:
                newkey = d[key]
    else:  # concepts, no need to change.
        newkey = ingredient.key
    if not result:
        result = ingredient.id + '-translated'
    return get_ingredient_class(ingredient.dtype).from_procedure_result(result, newkey, data_computed=new_data)



# -*- coding: utf-8 -*-

import os.path as osp

from functools import partial

# TODO: maybe move this module out of model/


# return absolute path from any path string
def absolute_path(path: str) -> str:
    if not osp.isabs(path):
        return osp.abspath(osp.expanduser(path))
    else:
        return path


# helper for sorting the json object
def sort_json(dp):
    """sort json object. dp means datapackage.json"""
    def get_sort_key(d, ks: list):
        res = []
        for k in ks:
            if isinstance(d[k], (list, tuple)):
                for v in d[k]:
                    res.append(v)
            elif d[k] is None:
                res.append('')
            else:
                res.append(d[k])
        return res

    def proc(x, ks):
        return sorted(x, key=partial(get_sort_key, ks=ks))

    if 'resources' in dp.keys():
        dp['resources'] = proc(dp['resources'], ['path'])

    if 'ddfSchema' in dp.keys():
        schema = dp['ddfSchema']
        for t in ['concepts', 'entities', 'datapoints']:
            if t in schema.keys():
                for v in schema[t]:
                    v['resources'] = sorted(v['resources'])
                schema[t] = proc(schema[t], ['value', 'primaryKey'])

        dp['ddfSchema'] = schema

    return dp





# -*- coding: utf-8 -*-
"""functions for datapackage.json"""

import os
import re
import json
import csv
import logging
from .model.package import Datapackage
from collections import OrderedDict


def get_datapackage(path, use_existing=True, update=True):
    """get the datapackage.json from a dataset path, create one if it's not exists

    Parameters
    ----------
    path : `str`
        the dataset path

    Keyword Args
    ------------
    use_existing : bool
        whether or not to use the existing datapackage
    """
    datapackage_path = os.path.join(path, 'datapackage.json')

    if os.path.exists(datapackage_path):
        with open(datapackage_path, encoding='utf8') as f:
            datapackage_old = json.load(f, object_pairs_hook=OrderedDict)

        if use_existing:
            if not update:
                return datapackage_old
            try:
                datapackage_old.pop('resources')  # don't use the old resources
                datapackage_old.pop('ddfSchema')  # and ddf schema
            except KeyError:
                logging.warning('no resources or ddfSchema in datapackage.json')
            datapackage_new = create_datapackage(path, **datapackage_old)
        else:
            datapackage_new = create_datapackage(path)
    else:
        if use_existing:
            print("WARNING: no existing datapackage.json")
        datapackage_new = create_datapackage(path)

    return datapackage_new


def get_ddf_files(path, root=None):
    info = next(os.walk(path))

    # don't include hidden and lang/etl dir.
    sub_dirs = [
        x for x in info[1] if (not x.startswith('.') and x not in ['lang', 'etl', 'langsplit'])
    ]
    files = list()
    for x in info[2]:
        if x.startswith('ddf--') and x != 'ddf--index.csv' and x.endswith('.csv'):
            files.append(x)
        else:
            logging.warning('skipping file {}'.format(x))

    for f in files:
        if root:
            yield os.path.join(root, f)
        else:
            yield f

    for sd in sub_dirs:
        for p in get_ddf_files(os.path.join(path, sd), root=sd):
            yield p


def create_datapackage(path, gen_schema=True, **kwargs):
    """create datapackage.json base on the files in `path`.

    If you want to set some attributes manually, you can pass them as
    keyword arguments to this function

    Note
    ----
    A DDFcsv datapackage MUST contain the fields `name` and `resources`.

    if name is not provided, then the base name of `path` will be used.

    Parameters
    ----------
    path : `str`
        the dataset path to create datapackage.json
    """

    datapackage = OrderedDict()

    # setting default name / lang
    try:
        name = kwargs.pop('name')
    except KeyError:
        # print('name not specified, using the path name')
        name = os.path.basename(os.path.normpath(os.path.abspath(path)))
    try:
        lang = kwargs.pop('language')
    except KeyError:
        lang = {'id': 'en'}

    datapackage['name'] = name
    datapackage['language'] = lang

    # add all optional settings
    for k in sorted(kwargs.keys()):
        datapackage[k] = kwargs[k]

    # generate resources
    resources = []
    names_sofar = dict()

    for f in get_ddf_files(path):
        path_res = f
        name_res = os.path.splitext(os.path.basename(f))[0]

        if name_res in names_sofar.keys():
            names_sofar[name_res] = names_sofar[name_res] + 1
            # adding a tail to the recource name, because it should be unique
            name_res = name_res + '-' + str(names_sofar[name_res])
        else:
            names_sofar[name_res] = 0

        resources.append(OrderedDict([('path', path_res), ('name', name_res)]))

    # TODO: make separate functions. this function is too long.
    for n, r in enumerate(resources):
        name_res = r['name']
        schema = {"fields": [], "primaryKey": None}

        if 'datapoints' in name_res:
            conc, keys = re.match('ddf--datapoints--([\w_]+)--by--(.*)', name_res).groups()
            primary_keys = keys.split('--')
            # print(conc, primary_keys)
            for i, k in enumerate(primary_keys):
                if '-' in k:
                    k_new, *enums = k.split('-')
                    primary_keys[i] = k_new
                    constraint = {'enum': enums}
                    schema['fields'].append({'name': k_new, 'constraints': constraint})
                else:
                    schema['fields'].append({'name': k})

            schema['fields'].append({'name': conc})
            schema['primaryKey'] = primary_keys

            resources[n].update({'schema': schema})

        elif 'entities' in name_res:
            match = re.match('ddf--entities--([\w_]+)(--[\w_]*)?-?.*', name_res).groups()
            domain, concept = match
            if concept is not None:
                concept = concept[2:]

            with open(os.path.join(path, r['path'])) as f:
                reader = csv.reader(f, delimiter=',', quotechar='"')
                # we only need the headers for index file
                header = next(reader)

            if domain in header:
                key = domain
            elif concept is not None and concept in header:
                key = concept
            else:
                raise ValueError('no header in {} matches its implied domain/entity_set!'.format(name_res))
                # print(
                #     """There is no matching header found for {}. Using the first column header
                #     """.format(name_res)
                # )
                # key = header[0]

            schema['primaryKey'] = key
            for h in header:
                schema['fields'].append({'name': h})
            resources[n].update({'schema': schema})

        elif 'concepts' in name_res:
            with open(os.path.join(path, r['path'])) as f:
                reader = csv.reader(f, delimiter=',', quotechar='"')
                header = next(reader)
            schema['primaryKey'] = 'concept'
            for h in header:
                schema['fields'].append({'name': h})

            resources[n].update({'schema': schema})
        else:  # not entity/concept/datapoint. it's not supported yet so we don't include them.
            print("not supported file: " + name_res)
            resources[n] = None

    datapackage['resources'] = [x for x in resources if x is not None]

    # generate ddf schema
    if gen_schema:
        dp = Datapackage(datapackage, base_dir=path)
        logging.info('generating ddf schema, may take some time...')
        dp.generate_ddfschema()

        return dp.datapackage
    else:
        return datapackage


# helper for dumping datapackage json
def dump_json(path, obj):
    with open(path, 'w+') as f:
        json.dump(obj, f, ensure_ascii=False, indent=4)
        f.close()



# -*- coding: utf-8 -*-

"""download sources from CME info portal

source link: `CME data portal`_

.. _`CME data portal`: http://www.childmortality.org

Warning
-------

This factory class is no longer works as CME website
switched to a javascript app.

"""


import os.path as osp

import re
import requests
import pandas as pd

from lxml import html
from urllib.parse import urlsplit, urljoin

from . common import DataFactory


class IGMELoader(DataFactory):

    url = 'https://www.childmortality.org/'

    def load_metadata(self):
        r = requests.get(self.url)
        h = html.fromstring(r.content)

        flist = []

        for l in h.xpath('//a/@href'):
            if l.endswith('xlsx'):
                # print(urljoin(url, l))
                flist.append(urljoin(self.url, l))

        md = pd.DataFrame(flist, columns=['link'])
        md['name'] = md['link'].map(lambda x: osp.basename(x)[:-5])

        metadata = md[['name', 'link']].copy()
        self.metadata = metadata
        return metadata

    def has_newer_source(self, v):
        """accepts a int and return true if version inferred from metadata is bigger."""
        if self.metadata is None:
            self.load_metadata()
        metadata = self.metadata
        link = metadata.loc[0, 'link']

        ver = re.match(r'.*files_v(\d+).*', link).groups()[0]

        if int(ver) > v:
            return True
        return False

    def bulk_download(self, out_dir, name=None):
        if self.metadata is None:
            self.load_metadata()
        metadata = self.metadata

        if name:
            names = [name]
        else:
            names = metadata['name'].values

        for n in names:
            if n not in metadata['name'].values:
                raise KeyError("{} not found in page.".format(n))

            link = metadata.loc[metadata['name'] == n, 'link'].values[0]
            res = requests.get(link)
            out_path = osp.join(out_dir, osp.basename(link))

            with open(osp.expanduser(out_path), 'wb') as f:
                f.write(res.content)
                f.close()



# -*- coding: utf-8 -*-

"""functions for common tasks on ddf datasets"""

import pandas as pd
import dask.dataframe as dd
import numpy as np
import json
import logging

from functools import partial
from ddf_utils.chef.helpers import prompt_select


logger = logging.getLogger(__name__)


def _check_ambiguities(dictionary, ambiguity):
    # check for ambiguities
    dict_ = dictionary.copy()
    for k, v in dictionary.items():
        if isinstance(v, list):  # there is ambiguity
            if ambiguity == 'skip':
                dict_.pop(k)
            elif ambiguity == 'error':
                raise ValueError("ambiguities found in dict_!")
            elif ambiguity == 'prompt':
                # prompt for value
                text = 'Please choose the correct entity to align *{}* with ' \
                       'for the rest of the execution of the recipe:'.format(k)
                val = prompt_select(v, text_before=text)
                if val != -1:
                    dict_[k] = val
                else:
                    dict_.pop(k)
    return dict_


def __print_not_found(series, dictionary):
    nf = set()
    if type(series) is dd.Series:
        uniques = series.unique().compute()
    else:
        uniques = series.unique()
    for v in uniques:
        if v not in dictionary.keys():
            nf.add(v)
    if len(nf) > 0:
        logger.warning('key not found:')
        logger.warning(list(nf))


def __process_val(v, dictionary=None):
    # check if a value is in the dictionary, if not, return nan
    # and mark down the not found values
    #
    # There is an issue in dask, see dask issue #4127
    # if v == 'foo':
    #     import ipdb; ipdb.set_trace()
    if v in dictionary.keys():
        return dictionary[v]
    else:
        return np.nan


def _translate_column_inline(df, column, target_column, dictionary,
                             not_found, ambiguity, ignore_case=False):
    df_new = df.copy()

    if ignore_case:
        df_new[column] = df_new[column].map(
            lambda x: str(x).lower() if x is not None else x)

    dictionary = _check_ambiguities(dictionary, ambiguity)

    # TODO: refactor this block
    # handling values not found in the dictionary
    if not_found == 'drop':
        __print_not_found(df_new[column], dictionary)
        process_val = partial(__process_val, dictionary=dictionary)
        df_new[target_column] = df_new[column].map(process_val)
        df_new = df_new.dropna(subset=[target_column])
    if not_found == 'error':
        df_new[target_column] = df_new[column].map(
            lambda x: dictionary[x])
    if not_found == 'include':
        if target_column not in df_new.columns:
            # create a new column: if a key not in the mappings, return NaN
            df_new[target_column] = df_new[column].map(
                lambda x: dictionary[x] if x in dictionary.keys() else np.nan)
        else:
            # update existing column: if a key not in the mappings, use the original val
            df_new['__new_col'] = df_new[column].map(
                lambda x: dictionary[x] if x in dictionary.keys() else np.nan)
            df_new['__new_col'] = df_new['__new_col'].fillna(df_new[target_column])
            df_new[target_column] = df_new['__new_col']
            df_new = df_new.drop('__new_col', axis=1)

    return df_new


def _generate_mapping_dict1(df, column, dictionary, base_df, not_found):

    search_col = dictionary['key']
    idx_col = dictionary['value']

    # Note: don't find mappings for n/a
    base_df_ = base_df.dropna(subset=[idx_col]).dropna(subset=[search_col])

    if base_df_.set_index(idx_col).index.has_duplicates:
        logger.warning('there are duplicated keys in the base dataframe:')
        m = base_df_.set_index(idx_col).index.duplicated()
        logger.warning(base_df_.set_index(idx_col).index[m].unique())

    if search_col == idx_col:
        mapping_all = dict([(x, x) for x in base_df_[idx_col].values])
    else:
        mapping_all = base_df_.set_index(search_col)[idx_col].to_dict()
    # add logger for no match
    no_match = set()

    # need to compute the actual dataframe now if it's a dask dataframe.
    # otherwise it will raise error.
    if isinstance(df, dd.DataFrame):
        uvs = df[column].compute().unique()
    else:
        uvs = df[column].unique()
    for v in uvs:
        if v not in mapping_all.keys():
            no_match.add(v)
    if len(no_match) > 0:
        logger.warning('no match found: ' + str(no_match))
    return mapping_all


def _generate_mapping_dict2(df, column, dictionary, base_df, not_found, ignore_case):

    mapping = dict()
    no_match = set()

    search_cols = dictionary['key']
    idx_col = dictionary['value']

    assert isinstance(idx_col, str)

    for f in df[column].unique():
        if pd.isnull(f):
            logger.warning('skipping n/a values in column {}'.format(column))
            continue
        bools = []
        for sc in search_cols:
            if ignore_case:
                bools.append(base_df[sc].str.lower() == f.lower())
            else:
                bools.append(base_df[sc] == f)

        mask = bools[0]
        for m in bools[1:]:
            mask = mask | m
        filtered = base_df[mask]

        if len(filtered) == 0 or np.all(pd.isnull(filtered[idx_col])):
            no_match.add(f)
            continue

        if len(filtered) == 1:
            mapping[f] = filtered[idx_col].iloc[0]
        elif len(filtered) > 1:
            logger.warning("multiple match found: "+f)
            mapping[f] = filtered[idx_col].values.tolist()

    if len(no_match) > 0:
        logger.warning('no match found: ' + str(no_match))

    if not_found == 'error' and len(no_match) > 0:
        raise ValueError('missing keys in dictionary. please check your input.')
    else:
        return mapping


def _translate_column_df(df, column, target_column, dictionary, base_df,
                         not_found, ambiguity, ignore_case):

    search_cols = dictionary['key']
    if isinstance(search_cols, str):
        mapping = _generate_mapping_dict1(df, column, dictionary, base_df, not_found)
    else:
        if len(search_cols) == 1:
            dictionary['key'] = search_cols[0]
            mapping = _generate_mapping_dict1(df, column, dictionary, base_df, not_found,
                                              ignore_case)
        else:
            mapping = _generate_mapping_dict2(df, column, dictionary, base_df, not_found,
                                              ignore_case)

    return _translate_column_inline(df, column, target_column, mapping, not_found, ambiguity)


def translate_column(df, column, dictionary_type, dictionary,
                     target_column=None, base_df=None, not_found='drop', ambiguity='prompt',
                     ignore_case=False):
    """change values in a column base on a mapping dictionary.

    The dictionary can be provided as a python dictionary, pandas dataframe or read from file.

    Note
    ----
    When translating with a base DataFrame, if ambiguity is found in the data, for example,
    a dataset with entity-id `congo`, to align to a dataset with `cod` ( Democratic Republic
    of the Congo ) and `cog` ( Republic of the Congo ), the function will ask for user input
    to choose which one or to skip it.

    Parameters
    ----------
    df : `DataFrame`
        The dataframe to be translated
    column : `str`
        The column to be translated
    dictionary_type : `str`
        The type of dictionary, choose from `inline`, `file` and `dataframe`
    dictionary : `str` or `dict`
        The dictionary. Depanding on the `dictionary_type`, the value of this parameter should be:
        `inline`: `dict`
        `file`: the file path, `str`
        `dataframe`: `dict`, must have `key` and `value` keys. see examples in examples section.
    target_column : `str`, optional
        The column to store translated resluts. If this is None, then the one set with `column`
        will be replaced.
    `base_df` : `DataFrame`, optional
        When `dictionary_type` is `dataframe`, this option should be set
    `not_found` : `str`
        What to do if key in the dictionary is not found in the dataframe to be translated.
        avaliable options are `drop`, `error`, `include`
    `ambiguity` : `str`
        What to do when there is ambiguities in the dictionary. avaliable options are `prompt`,
        `skip`, `error`

    Examples
    --------
    >>> df = pd.DataFrame([['geo', 'Geographical places'], ['time', 'Year']], columns=['concept', 'name'])
    >>> df
      concept                 name
    0     geo  Geographical places
    1    time                 Year
    >>> translate_column(df, 'concept', 'inline', {'geo': 'country', 'time': 'year'})
       concept                 name
    0  country  Geographical places
    1     year                 Year

    >>> base_df = pd.DataFrame([['geo', 'country'], ['time', 'year']], columns=['concept', 'alternative_name'])
    >>> base_df
      concept alternative_name
    0     geo          country
    1    time             year
    >>> translate_column(df, 'concept', 'dataframe',
    ...                     {'key': 'concept', 'value': 'alternative_name'},
    ...                     target_column='new_name', base_df=base_df)
      concept                 name new_name
    0     geo  Geographical places  country
    1    time                 Year     year
    """

    if dictionary_type not in ['inline', 'file', 'dataframe']:
        raise ValueError("dictionary_type should be one of 'inline', 'file', 'dataframe'")

    if not_found not in ['drop', 'error', 'include']:
        raise ValueError("not_found should be one of 'drop', 'error', 'include'")

    if ambiguity not in ['skip', 'error', 'prompt']:
        raise ValueError("ambiguity should be one of 'skip', 'error', 'prompt'")

    if dictionary_type == 'dataframe' and base_df is None:
        raise ValueError("please specify base_df when dictionary type is 'dataframe'")

    if not target_column:
        target_column = column

    if dictionary_type == 'inline':
        df_new = _translate_column_inline(df, column, target_column, dictionary,
                                          not_found, ambiguity, ignore_case)
    elif dictionary_type == 'file':
        with open(dictionary) as f:
            d = json.load(f)
        df_new = _translate_column_inline(df, column, target_column, d,
                                          not_found, ambiguity)
    elif dictionary_type == 'dataframe':
        assert 'key' in dictionary.keys() and 'value' in dictionary.keys()
        df_new = _translate_column_df(df, column, target_column, dictionary, base_df,
                                      not_found, ambiguity, ignore_case)
    else:
        raise ValueError("dictionary_type not supported: ()".format(dictionary_type))

    return df_new


def translate_header(df, dictionary, dictionary_type='inline'):
    """change the headers of a dataframe base on a mapping dictionary.

    Parameters
    ----------
    df : `DataFrame`
        The dataframe to be translated
    dictionary_type : `str`, default to `inline`
        The type of dictionary, choose from `inline` or `file`
    dictionary : `dict` or `str`
        The mapping dictionary or path of mapping file
    """
    if dictionary_type == 'inline':
        return df.rename(columns=dictionary)
    elif dictionary_type == 'file':
        with open(dictionary, 'r') as f:
            dictionary = json.load(f)
        return df.rename(columns=dictionary)
    else:
        raise ValueError('dictionary not supported: '+dictionary_type)


def trend_bridge(old_ser: pd.Series, new_ser: pd.Series, bridge_length: int) -> pd.Series:
    """smoothing data between series.

    To avoid getting artificial stairs in the data, we smooth between to
    series. Sometime one source is systematically higher than another source,
    and if we jump from one to another in a single year, this looks like an
    actual change in the data.

    Parameters
    ----------
    old_data : Series
    new_data : Series
    bridge_length : int
        the length of bridge

    Returns
    -------
    bridge_data : the bridged data

    """
    old_data = old_ser.sort_index()
    new_data = new_ser.sort_index()

    if old_data.empty:
        return new_data
    if new_data.empty:
        return old_data

    bridge_end = new_data.index[0]

    if old_data.index[0] > bridge_end:  # not bridging in this case
        return new_data

    if old_data.index[-1] < bridge_end:
        return pd.concat([old_data, new_data], sort=False)

    # recifying bridge_end, if old_data do not have data at this point
    if bridge_end not in old_data.index:
        intersection = old_data.loc[old_data.index.isin(new_data.index)]
        if intersection.empty:
            raise ValueError("can't bridge because there is no intersection time point")
        bridge_end = intersection.index[0]

    # now recify bridge_length/bridge_start, because in some case old_data just don't have enough data
    s1 = old_data.loc[:bridge_end]
    if s1.shape[0] < bridge_length:
        bridge_length = s1.shape[0]
        bridge_start = old_data.index[0]
    else:
        bridge_start = s1.index.values[-bridge_length:][0]

    bridge_height = new_data.loc[bridge_end] - old_data.loc[bridge_end]
    fraction = bridge_height / bridge_length

    bridge_data = old_data.copy()

    for i in old_data.loc[bridge_start:bridge_end].index:
        bridge_data.loc[i:bridge_end] = bridge_data.loc[i:bridge_end] + fraction

    # combine old/new/bridged data
    result = pd.concat([bridge_data.loc[:bridge_end], new_data.loc[bridge_end:].iloc[1:]], sort=False)
    return result


def extract_concepts(dfs, base=None, join='full_outer'):
    """extract concepts from a list of dataframes.

    Parameters
    ----------
    dfs : list[DataFrame]
        a list of dataframes to be extracted

    Keyword Args
    ------------
    base : DataFrame
        the base concept table to join
    join : {'full_outer', 'ingredients_outer'}
        how to join the `base` dataframe. ``full_outer`` means union of the base and extracted,
        ``ingredients_outer`` means only keep concepts in extracted

    Return
    ------
    DataFrame
        the result concept table
    """
    if base is not None:
        concepts = base.set_index('concept')
    else:
        concepts = pd.DataFrame([], columns=['concept', 'concept_type']).set_index('concept')

    new_concepts = set()

    for df in dfs:
        for c in df.columns:
            new_concepts.add(c)
            if c in concepts.index:  # if the concept is in base, just use base data
                continue
            if c.startswith('is--'):
                ent = c[4:]
                new_concepts.add(ent)
                # concept in the is--header should be entity_set
                concepts.loc[c[4:], 'concept_type'] = 'entity_set'
                continue
            if np.issubdtype(df[c].dtype, np.number):
                concepts.loc[c, 'concept_type'] = 'measure'
            else:
                concepts.loc[c, 'concept_type'] = 'string'
    if join == 'ingredients_outer':
        # ingredients_outer join: only keep concepts appears in ingredients
        concepts = concepts.loc[new_concepts]
    return concepts.reset_index()


def merge_keys(df, dictionary, target_column=None, merged='drop'):
    """merge keys"""
    rename_dict = dict()
    for new_key, val in dictionary.items():
        for old_key in val:
            rename_dict[old_key] = new_key
    # TODO: limit the rename inside target_column
    # after pandas 0.20.0 there will be a level option for df.rename
    df_new = df.rename(index=rename_dict).groupby(level=list(range(len(df.index.levels)))).sum()
    if merged == 'drop':
        return df_new
    elif merged == 'keep':
        df_ = pd.concat([df, df_new])
        return df_[~df_.index.duplicated()]  # remove all duplicated indies
    else:
        raise ValueError('only "drop", "keep" is allowed')


def split_keys(df, target_column, dictionary, splited='drop'):
    """split entities"""
    keys = df.index.names
    df_ = df.reset_index()

    ratio = dict()

    for k, v in dictionary.items():
        # dictionary format:
        # {entity_to_split: [sub_entity_1, ...]}
        # the split ratio will be calculated from first valid values of sub entities.
        # so it assumes existence of sub entities
        # TODO: maybe make it work even sub entities not exists.
        before_spl = list()
        for spl in v:
            if spl not in df_[target_column].values:
                raise ValueError('entity not in data: ' + spl)
            tdf = df_[df_[target_column] == spl].set_index(keys).sort_index()
            last = pd.DataFrame(tdf.loc[tdf.index[0], tdf.columns]).T
            last.index.names = keys
            logger.debug("using {} for first valid index".format(tdf.index[0]))
            last = last.reset_index()
            for key in keys:
                if key != target_column:
                    last = last.drop(key, axis=1)
            before_spl.append(last.set_index(target_column))
        before_spl = pd.concat(before_spl)
        total = before_spl.sum()
        ptc = before_spl / total
        ratio[k] = ptc.to_dict()

    logger.debug(ratio)
    # the ratio format will be:
    # ratio = {
    #     'entity_to_split': {
    #         'concept_1': {
    #             'sub_entity_1': r11,
    #             'sub_entity_2': r12,
    #             ...
    #         },
    #         'concept_2': {
    #             'sub_entity_1': r21,
    #             'sub_entity_2': r22,
    #             ...
    #         }
    #     }
    # }

    to_concat = []
    for k, v in ratio.items():
        t = df_[df_[target_column] == k].copy()
        for x, y in v.items():
            for g, r in y.items():
                t_ = t.copy()
                t_[x] = t_[x] * r
                t_[target_column] = g
                to_concat.append(t_.set_index(keys))

    to_concat.append(df)
    final = pd.concat(to_concat)
    if splited == 'drop':
        final = final[~final.index.get_level_values(target_column).isin(dictionary.keys())]
        return final.sort_index()
    elif splited == 'keep':
        return final.sort_index()
    else:
        raise ValueError('only support drop == "drop" and "keep".')



# -*- coding: utf-8 -*-
"""i18n project management for Gapminder's datasets.

The workflow is described in this `google doc`_. The json part comes from discussion `here`_

.. _google doc: https://docs.google.com/document/d/11d5D5CPlr6I2BqP8z0p2o_dYQC9AYUyxcLGUq1WYCfk/
.. _here: https://gapminder.slack.com/archives/C0FTBA1P0/p1479832399000650

"""

import os
import json
import numpy as np
import pandas as pd
from .package import get_datapackage


def split_translations_json(path, split_path='langsplit', exclude_concepts=None, overwrite=False):
    """split all string concepts and save them as json files

    Note
    ----

    There is an issue with dataframe.to_json() method for multiIndex files (i.e. datapoints), which
    cause we can't read back the split files and merge them. In this case :py:func:`merge_translations_json` will fail.

    """
    datapackage = get_datapackage(path)
    split_path = os.path.join(path, split_path)

    try:
        lang = datapackage['language']['id']
    except KeyError:
        print('no language tag found, assuming en')
        lang = 'en'

    basepath = os.path.join(split_path, lang)
    os.makedirs(basepath, exist_ok=True)

    # exclude some columns for translation by default
    if not exclude_concepts:
        exclude_concepts = ['concept_type', 'domain']

    concepts = list()
    for res in datapackage['resources']:
        file_path = res['path']
        key = res['schema']['primaryKey']
        if key == 'concept':
            concepts.append(pd.read_csv(os.path.join(path, file_path)))
    concepts = pd.concat(concepts, ignore_index=True).set_index('concept')

    for res in datapackage['resources']:
        file_path = res['path']
        key = res['schema']['primaryKey']

        df = pd.read_csv(os.path.join(path, file_path))
        df = df.set_index(key)
        for c in list(df.columns):
            if c in exclude_concepts:
                df = df.drop(c, axis=1)
                continue
            if c.startswith('is--'):  # it will be boolean, skip
                df = df.drop(c, axis=1)
                continue
            try:
                if not concepts.loc[c, 'concept_type'] == 'string':
                    df = df.drop(c, axis=1)
                    continue
            except KeyError:
                print('concept not found in ddf--concepts: ' + c)
                df = df.drop(c, axis=1)
                continue
        # save file to disk
        if not np.all(pd.isnull(df)):
            json_path = os.path.join(basepath, res['name']+'.json')
            dir_path = os.path.dirname(json_path)
            if os.path.exists(json_path) and not overwrite:
                print('file exists: ' + json_path)
                continue
            os.makedirs(dir_path, exist_ok=True)
            # print(json_path)
            df = df.fillna('')
            # FIXME: handle multiIndex issue
            # see https://github.com/pandas-dev/pandas/issues/4889
            df.to_json(json_path, orient='index', force_ascii=False)
    return


def split_translations_csv(path, split_path='langsplit', exclude_concepts=None, overwrite=False):
    """split all string concepts and save them as csv files"""
    datapackage = get_datapackage(path)
    split_path = os.path.join(path, split_path)

    try:
        lang = datapackage['language']['id']
    except KeyError:
        print('no language tag found, assuming en')
        lang = 'en'

    basepath = os.path.join(split_path, lang)
    os.makedirs(basepath, exist_ok=True)

    # exclude some columns for translation by default
    if not exclude_concepts:
        exclude_concepts = ['concept', 'concept_type', 'domain']

    concepts = list()
    for res in datapackage['resources']:
        file_path = res['path']
        key = res['schema']['primaryKey']
        if key == 'concept':
            concepts.append(pd.read_csv(os.path.join(path, file_path)))
    concepts = pd.concat(concepts, ignore_index=True).set_index('concept')

    for res in datapackage['resources']:
        file_path = res['path']
        key = res['schema']['primaryKey']

        df = pd.read_csv(os.path.join(path, file_path))
        for c in df.columns:
            if c in exclude_concepts:
                continue
            if c.startswith('is--'):  # it will be boolean, skip
                continue
            try:
                if concepts.loc[c, 'concept_type'] == 'string':
                    os.makedirs(os.path.join(basepath, file_path), exist_ok=True)
                    split_csv_path = os.path.join(basepath, file_path, '{}.csv'.format(c))
                    if os.path.exists(split_csv_path) and not overwrite:
                        print('file exists: ' + split_csv_path)
                        continue
                    df.set_index(key)[[c]].to_csv(split_csv_path, encoding='utf8')
            except KeyError:
                print('concept not found in ddf--concepts: ' + c)
                continue
    return


def merge_translations_csv(path, split_path='langsplit', lang_path='lang', overwrite=False):
    """merge all translated csv files and update datapackage.json"""
    if overwrite:
        # TODO: overwrite existing translation instead of update
        raise NotImplementedError

    # make the paths full paths
    split_path = os.path.join(path, split_path)
    lang_path = os.path.join(path, lang_path)

    datapackage = get_datapackage(path)
    source_lang = datapackage['language']['id']
    # get all available translations, update it when necessary below.
    if 'translations' not in datapackage.keys():
        datapackage['translations'] = []
    available_translations = [x['id'] for x in datapackage['translations']]
    new_translations = False

    assert os.path.exists(split_path)

    langs = next(os.walk(split_path))[1]  # all sub folders in split path.

    for lang in langs:
        if lang == source_lang:
            continue

        if lang not in available_translations:
            datapackage['translations'].append({'id': lang})
            new_translations = True

        basepath = os.path.join(split_path, lang) + '/'
        target_lang_path = os.path.join(lang_path, lang)
        os.makedirs(target_lang_path, exist_ok=True)

        for root, dirs, files in os.walk(basepath):
            csvfiles = [x for x in files if 'csv' in x]
            if len(csvfiles) > 0:  # there are csv files.
                p = root.replace(basepath, '')

                os.makedirs(os.path.join(target_lang_path, os.path.dirname(p)), exist_ok=True)

                target_file_path = os.path.join(target_lang_path, p)
                to_concat = [pd.read_csv(os.path.join(root, x), index_col=0) for x in csvfiles]
                df_new = pd.concat(to_concat, axis=1)
                if os.path.exists(target_file_path):
                    df_old = pd.read_csv(target_file_path, index_col=0)
                    df_old.update(df_new)
                    df_new = df_old.copy()
                df_new.to_csv(target_file_path, encoding='utf8')
    # update datapackage if there are new translations
    if new_translations:
        with open(os.path.join(path, 'datapackage.json'), 'w') as f:
            json.dump(datapackage, f, indent=4, ensure_ascii=False)
    return


def merge_translations_json(path, split_path='langsplit', lang_path='lang', overwrite=False):
    """merge all translated json files and update datapackage.json.
    """
    # make the paths full paths
    split_path = os.path.join(path, split_path)
    lang_path = os.path.join(path, lang_path)

    datapackage = get_datapackage(path)
    source_lang = datapackage['language']['id']

    # because the primaryKey is not in the splited files, we generate
    # a mapping for later use.
    key_mapping = dict([(x['name'], x['schema']['primaryKey']) for x in datapackage['resources']])

    # get all available translations, update it when necessary below.
    if 'translations' not in datapackage.keys():
        datapackage['translations'] = []
    available_translations = [x['id'] for x in datapackage['translations']]
    new_translations = False

    assert os.path.exists(split_path)

    langs = next(os.walk(split_path))[1]  # all sub folders in split path.

    for lang in langs:
        if lang == source_lang:
            continue

        if lang not in available_translations:
            datapackage['translations'].append({'id': lang})
            new_translations = True

        basepath = os.path.join(split_path, lang) + '/'
        target_lang_path = os.path.join(lang_path, lang)
        os.makedirs(target_lang_path, exist_ok=True)

        for root, dirs, files in os.walk(basepath):
            jsonfiles = [x for x in files if 'json' in x]
            if len(jsonfiles) > 0:
                p = root.replace(basepath, '')
                for fn in jsonfiles:
                    name = os.path.splitext(fn)[0]
                    target_file_path = os.path.join(target_lang_path, p, fn).replace('.json', '.csv')
                    if not os.path.exists(os.path.join(target_lang_path, p)):
                        os.makedirs(os.path.join(target_lang_path, p))
                    df = pd.read_json(os.path.join(root, fn), orient='index')
                    # add back the index name.
                    # the index name will be missing because it's not in the json file.
                    df.index.name = key_mapping[name]
                    if os.path.exists(target_file_path) and not overwrite:
                        df_old = pd.read_csv(target_file_path, index_col=0)
                        df_old.update(df)
                        df = df_old.copy()
                    # dropping lines without any translations
                    df = df.dropna(how='all', axis=1)
                    df.to_csv(target_file_path, encoding='utf8')
    # update datapackage if there are new translations
    if new_translations:
        with open(os.path.join(path, 'datapackage.json'), 'w') as f:
            json.dump(datapackage, f, indent=4, ensure_ascii=False)
    return



# -*- coding: utf-8 -*-

"""groupby procedure for recipes"""

import fnmatch
import logging

from typing import List

from .. helpers import debuggable, mkfunc
from .. model.ingredient import DataPointIngredient
from .. model.chef import Chef


logger = logging.getLogger('groupby')


@debuggable
def groupby(chef: Chef, ingredients: List[DataPointIngredient], result, **options) -> DataPointIngredient:
    """group ingredient data by column(s) and run aggregate function

    .. highlight:: yaml

    Procedure format:

    ::

       procedure: groupby
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingredient id
       options:
         groupby: str or list  # column(s) to group
         aggregate: dict  # function block
         transform: dict  # function block
         filter: dict  # function block

    The function block should have below format:

    ::

      aggregate:
        column1: func_name1
        column2: func_name2

    or

    ::

        aggrgrate:
          column1:
            function: func_name
            param1: foo
            param2: baz

    wildcard is supported in the column names. So ``aggreagte: {"*": "sum"}`` will run on every indicator in
    the ingredient

    Keyword Args
    ------------
    groupby : `str` or `list`
        the column(s) to group, can be a list or a string
    insert_key : `dict`
        manually insert keys in to result. This is useful when we want to add back the
        aggregated column and set them to one value. For example ``geo: global`` inserts
        the ``geo`` column with all values are "global"
    aggregate
    transform
    filter : `dict`, optinoal
        the function to run. only one of `aggregate`, `transform` and `filter` should be supplied.

    Note
    ----
    - Only one of ``aggregate``, ``transform`` or ``filter`` can be used in one procedure.
    - Any columns not mentioned in groupby or functions are dropped.
    """
    assert len(ingredients) == 1, "procedure only support 1 ingredient for now."

    # ingredient = chef.dag.get_node(ingredients[0]).evaluate()
    ingredient = ingredients[0]
    logger.info("groupby: " + ingredient.id)
    data = ingredient.get_data()
    by = options.pop('groupby')
    if 'insert_key' in options:
        insert_key = options.pop('insert_key')
    else:
        insert_key = dict()

    # only one of aggregate/transform/filter should be in options.
    assert len(list(options.keys())) == 1
    comp_type = list(options.keys())[0]
    assert comp_type in ['aggregate', 'transform', 'filter']

    if comp_type == 'aggregate':  # only aggregate should change the key of ingredient
        if isinstance(by, list):
            newkey = ','.join(by)
        else:
            newkey = by
            by = [by]
        logger.debug("changing the key to: " + str(newkey))
    else:
        newkey = ingredient.key
        by = [by]

    newdata = dict()

    for name_tmpl, func in options[comp_type].items():
        func = mkfunc(func)
        indicator_names = fnmatch.filter(data.keys(), name_tmpl)
        for k in indicator_names:
            df = data[k].compute()
            if comp_type == 'aggregate':
                newdata[k] = (df.groupby(by=by).agg({k: func})
                              .reset_index().dropna())
            if comp_type == 'transform':
                df = df.set_index(ingredient.key)
                levels = [df.index.names.index(x) for x in by]
                newdata[k] = (df.groupby(level=levels)[k].transform(func)
                              .reset_index().dropna())
            if comp_type == 'filter':
                df = df.set_index(ingredient.key)
                levels = [df.index.names.index(x) for x in by]
                newdata[k] = (df.groupby(level=levels)[k].filter(func)
                              .reset_index().dropna())
            for col, val in insert_key.items():
                newdata[k][col] = val
                newkey = newkey + ',' + col

    return DataPointIngredient.from_procedure_result(result, newkey, data_computed=newdata)



# -*- coding: utf-8 -*-

"""all procedures for recipes"""

import pandas as pd
import numpy as np
from . ingredient import BaseIngredient, ProcedureResult
from .helpers import read_opt, mkfunc, debuggable
from .exceptions import ProcedureError
import time
from typing import List, Union, Dict, Optional

import logging

logger = logging.getLogger('Chef')


@debuggable
def translate_header(ingredient: BaseIngredient, result, dictionary) -> ProcedureResult:
    """Translate column headers

    Procedure format:

    .. code-block:: yaml

       procedure: translate_header
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingredient id
       options:
         dictionary: str or dict  # file name or mappings dictionary

    Parameters
    ----------
    ingredient : BaseIngredient
        The ingredient to translate
    result : `str`
        The result ingredient id

    Keyword Args
    ------------
    dictionary: dict
        a dictionary of oldname -> newname mappings

    See Also
    --------
    :py:func:`ddf_utils.transformer.translate_header` : Related function in transformer module
    """
    logger.info("translate_header: " + ingredient.ingred_id)

    rm = dictionary
    data = ingredient.copy_data()

    for k in list(data.keys()):
        df_new = data[k].rename(columns=rm).copy()
        if ingredient.dtype == 'entities':  # also rename the `is--` headers
            rm_ = {}
            for c in df_new.columns:
                if k == c[4:]:
                    rm_[c] = 'is--' + rm[k]
            if len(rm_) > 0:
                df_new = df_new.rename(columns=rm_)
        if k in rm.keys():  # if we need to rename the concept name
            data[rm[k]] = df_new
            del(data[k])
        else:  # we only rename index/properties columns
            data[k] = df_new

    # also rename the key
    newkey = ingredient.key
    if ingredient.dtype == 'datapoints':
        for key in rm.keys():
            if key in ingredient.key:
                newkey = newkey.replace(key, rm[key])
    elif ingredient.dtype == 'entities':
        for key in rm.keys():
            if key == ingredient.key:
                newkey = rm[key]
    else:
        if 'concept' in rm.keys():
            raise ValueError('can translate the primaryKey for concept!')

    if not result:
        result = ingredient.ingred_id + '-translated'
    return ProcedureResult(result, newkey, data=data)


@debuggable
def translate_column(ingredient: BaseIngredient, result, dictionary, column, *,
                     target_column=None, not_found='drop', ambiguity='prompt',
                     ignore_case=False) -> ProcedureResult:
    """Translate column values.

    Procedure format:

    .. code-block:: yaml

       procedure: translate_column
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingledient id
       options:
         column: str  # the column to be translated
         target_column: str  # optinoal, the target column to store the translated data
         not_found: {'drop', 'include', 'error'}  # optional, the behavior when there is values not
                                                  # found in the mapping dictionary, default is 'drop'
         ambiguity: {'prompt', 'skip', 'error'}  # optional, the behavior when there is ambiguity
                                                 # in the dictionary
         dictionary: str or dict  # file name or mappings dictionary

    If base is provided in dictionary, key and value should also in dictionary.
    In this case chef will generate a mapping dictionary using the base ingredient.
    The dictionary format will be:

    .. code-block:: yaml

       dictionary:
         base: str  # ingredient name
         key: str or list  # the columns to be the keys of the dictionary, can accept a list
         value: str  # the column to be the values of the the dictionary, must be one column

    Keyword Args
    ------------
    dictionary: dict
        A dictionary of oldname -> newname mappings.
        If 'base' is provided in the dictionary, 'key' and 'value' should also in the dictionary.
        See :py:func:`ddf_utils.transformer.translate_column` for more on how this is handled.
    column: `str`
        the column to be translated
    target_column : `str`, optional
        the target column to store the translated data. If this is not set then the `column`
        cloumn will be replaced
    not_found : {'drop', 'include', 'error'}, optional
        the behavior when there is values not found in the mapping dictionary, default is 'drop'
    ambiguity : {'prompt', 'skip', 'error'}, optional
        the behavior when there is ambiguity in the dictionary, default is 'prompt'

    See Also
    --------
    :py:func:`ddf_utils.transformer.translate_column` : related function in transformer module
    """
    logger.info("translate_column: " + ingredient.ingred_id)

    from ..transformer import translate_column as tc

    di = ingredient.copy_data()

    # find out the type of dictionary.
    if isinstance(dictionary, str):
        dict_type = 'file'
        base_df = None
    else:
        if 'base' in dictionary.keys():
            dict_type = 'dataframe'
            base = dictionary.pop('base')
            base_data = base.get_data()
            if len(base_data) > 1:
                raise ProcedureError('only support ingredient with 1 item')
            base_df = list(base_data.values())[0]
        else:
            dict_type = 'inline'
            base_df = None

    for k, df in di.items():
        logger.debug("running on: " + k)
        di[k] = tc(df, column, dict_type, dictionary, target_column, base_df,
                   not_found, ambiguity, ignore_case)

    if not result:
        result = ingredient.ingred_id + '-translated'
    return ProcedureResult(result, ingredient.key, data=di)


@debuggable
def copy(ingredient: BaseIngredient, result, dictionary: Dict) -> ProcedureResult:
    """make copy of ingredient data columns, with new names.

    Procedure format:

    .. code-block:: yaml

       procedure: copy
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingledient id
       options:
       dictionary: dict  # old name -> new name mappings

    An example of dictionary:

    .. code-block:: yaml

       dictionary:
         col1: copy_1_1  # string
         col2:  # list of string
           - copy_2_1
           - copy_2_2

    where 'col1' and 'col2' should be existing columns in the input ingredient

    Keyword Args
    ------------
    dictionary: dict
        a dictionary of oldname -> newname mappings

    """
    logger.info("copy: " + ingredient.ingred_id)

    data = ingredient.copy_data()

    for k, v in dictionary.items():
        if isinstance(v, str):  # value is str, means only make one copy
            data[v] = data[k].rename(columns={k: v}).copy()
        else:  # then it's a list, should make multiple copy
            for n in v:
                data[n] = data[k].rename(columns={k: n}).copy()

    # usually the old ingredient won't be used after creating copy.
    # just reset the data to save memory
    ingredient.reset_data()
    if not result:
        result = ingredient.ingred_id + '_'
    return ProcedureResult(result, ingredient.key, data=data)


@debuggable
def merge(*ingredients: List[BaseIngredient], result, deep=False) -> ProcedureResult:
    """merge a list of ingredients

    The ingredients will be merged one by one in the order of how they are provided to this
    function. Later ones will overwrite the pervious merged results.

    Procedure format:

    .. code-block:: yaml

       procedure: merge
       ingredients:  # list of ingredient id
         - ingredient_id_1
         - ingredient_id_2
         - ingredient_id_3
         # ...
       result: str  # new ingledient id
       options:
         deep: bool  # use deep merge if true

    Parameters
    ----------
    BaseIngredient
        Any numbers of ingredients to be merged

    Keyword Args
    ------------
    deep: `bool`, optional
        if True, then do deep merging. Default is False

    Notes
    -----
    **deep merge** is when we check every datapoint for existence
    if false, overwrite is on the file level. If key-value
    (e.g. geo,year-population_total) exists, whole file gets overwritten
    if true, overwrite is on the row level. If values
    (e.g. afr,2015-population_total) exists, it gets overwritten, if it doesn't it stays
    """
    logger.info("merge: " + str([i.ingred_id for i in ingredients]))

    # assert that dtype and key are same in all dataframe
    try:
        for x in ingredients[1:]:
            assert set(x.key_to_list()) == set(ingredients[0].key_to_list())
        assert len(set([x.dtype for x in ingredients])) == 1
    except (AssertionError, TypeError):
        log1 = "multiple dtype/key detected: \n"
        log2 = "\n".join(["{}: {}, {}".format(x.ingred_id, x.dtype, x.key) for x in ingredients])
        logger.warning(log1+log2)
        raise ProcedureError("can't merge data with multiple dtype/key!")

    # get the dtype and index
    # we have assert dtype and key is unique, so we take it from
    # the first ingredient
    dtype = ingredients[0].dtype

    if dtype == 'datapoints':
        index_col = ingredients[0].key_to_list()
        newkey = ','.join(index_col)
    else:
        index_col = ingredients[0].key
        newkey = index_col

    if deep:
        logger.info("merge: doing deep merge")
    # merge data from ingredients one by one.
    res_all = {}

    for i in ingredients:
        res_all = _merge_two(res_all, i.get_data(), index_col, dtype, deep)

    if not result:
        result = 'all_data_merged_'+str(int(time.time() * 1000))

    return ProcedureResult(result, newkey, data=res_all)


def __get_last_item(ser):
    """get the last vaild item of a Series, or Nan."""
    if ser.last_valid_index() is None:
        return np.nan
    else:
        return ser[ser.last_valid_index()]


def _merge_two(left: Dict[str, pd.DataFrame],
               right: Dict[str, pd.DataFrame],
               index_col: Union[List, str],
               dtype: str, deep=False) -> Dict[str, pd.DataFrame]:
    """merge 2 ingredient data."""
    if len(left) == 0:
        return right

    if dtype == 'datapoints':
        if deep:
            for k, df in right.items():
                if k in left.keys():
                    left[k] = left[k].append(df, ignore_index=True)
                    left[k] = left[k].drop_duplicates(subset=index_col, keep='last')
                    left[k] = left[k].sort_values(by=index_col)
                else:
                    left[k] = df
        else:
            for k, df in right.items():
                left[k] = df

        res_data = left

    elif dtype == 'concepts':

        left_df = pd.concat(left.values())
        right_df = pd.concat(right.values())

        if deep:
            merged = left_df.append(right_df, ignore_index=True)
            res = merged.groupby(index_col).agg(__get_last_item)
            res_data = {'concept': res.reset_index()}
        else:
            res_data = {'concept': right_df.drop_duplicates(subset='concept', keep='last')}
    else:
        # TODO: improve this
        if deep:
            for k, df in right.items():
                if k in left.keys():
                    left[k] = left[k].append(df, ignore_index=True)
                    left[k] = left[k].groupby(index_col).agg(__get_last_item).reset_index()
                else:
                    left[k] = df
        else:
            for k, df in right.items():
                left[k] = df
        res_data = left
        # raise NotImplementedError('entity data do not support merging yet.')

    return res_data


@debuggable
def identity(ingredient: BaseIngredient, result, copy=False) -> BaseIngredient:
    """return the ingredient as is.

    Keyword Args
    ------------
    copy: bool, optional
        if copy is True, then treat all data as string. Default: False
    """
    if copy:
        ingredient.data = ingredient.get_data(copy=True)
    else:
        ingredient.data = ingredient.get_data()

    if result:
        ingredient.ingred_id = result + '-identity'
    return ingredient


@debuggable
def filter_row(ingredient: BaseIngredient, result, **options) -> ProcedureResult:
    """filter an ingredient based on a set of options and return
    the result as new ingredient.

    Procedure format:

    .. code-block:: yaml

       procedure: filter_row
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingledient id
       options:
         dictionary: dict  # filter definition block

    A dictionary should be provided in options with the following format:

    .. code-block:: yaml

        dictionary:
          new_key_in_new_ingredient:
            from: old_key_in_old_ingredient
            filter_col_1: filter_val_1
            filter_col_2: filter_val_2

    See a detail example in this `github issue <https://github.com/semio/ddf_utils/issues/2#issuecomment-254132615>`_

    Parameters
    ----------
    ingredient: BaseIngredient
    result: `str`

    Keyword Args
    ------------
    dictionary: dict
        The filter description dictionary
    keep_all_columns: bool
        don't drop any column if true
    """

    logger.info("filter_row: " + ingredient.ingred_id)

    data = ingredient.get_data()
    dictionary = read_opt(options, 'dictionary', True)
    keep_all_columns = read_opt(options, 'keep_all_columns', False, False)

    res = {}

    for k, v in dictionary.items():
        from_name = v.pop('from')
        df = data[from_name]
        if len(v) == 0:
            res[k] = df.rename(columns={from_name: k})
            continue
        queries = []

        for col, val in v.items():
            if isinstance(val, list):
                queries.append("{} in {}".format(col, val))
            elif isinstance(val, str):
                queries.append("{} == '{}'".format(col, val))
            elif np.issubdtype(type(val), np.number):
                queries.append("{} == {}".format(col, val))
            # TODO: support more query methods.
            else:
                raise ProcedureError("not supported in query: " + str(type(val)))

        query_string = ' and '.join(queries)

        df = df.query(query_string).copy()
        df = df.rename(columns={from_name: k})
        # drops all columns with unique contents. and update the key.
        newkey = ingredient.key
        keys = ingredient.key_to_list()
        if not keep_all_columns:
            for c in v.keys():
                if ingredient.dtype == 'datapoints':
                    if len(df[c].unique()) > 1:
                        logger.debug("column {} have multiple values: {}".format(c, df[c].unique()))
                    elif len(df[c].unique()) <= 1:
                        df = df.drop(c, axis=1)
                        if c in keys:
                            keys.remove(c)
                        newkey = ','.join(keys)
                else:
                    raise NotImplementedError("filtering concept/entity")

        res[k] = df

    if not result:
        result = ingredient.ingred_id + '-filtered'
    return ProcedureResult(result, newkey, data=res)


@debuggable
def filter_item(ingredient: BaseIngredient, result, items: list) -> ProcedureResult:
    """filter items from the ingredient data

    Procedure format:

    .. code-block:: yaml

       procedure: filter_item
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingledient id
       options:
         items: list  # a list of items should be in the result ingredient

    Keyword Args
    ------------
    items: list
        a list of items to filter from base ingredient
    """
    logger.info("filter_item: " + ingredient.ingred_id)

    data = ingredient.get_data()

    try:
        data = dict([(k, data[k]) for k in items])
    except KeyError as e:
        logger.debug("keys in {}: {}".format(ingredient.ingred_id, str(list(data.keys()))))
        raise ProcedureError(e.message)

    if not result:
        result = ingredient.ingred_id

    return ProcedureResult(result, ingredient.key, data=data)


@debuggable
def groupby(ingredient: BaseIngredient, result, **options) -> ProcedureResult:
    """group ingredient data by column(s) and run aggregate function

    .. highlight:: yaml

    Procedure format:

    ::

       procedure: groupby
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingledient id
       options:
         groupby: str or list  # colunm(s) to group
         aggregate: dict  # function block
         transform: dict  # function block
         filter: dict  # function block

    The function block should have below format:

    ::

      aggregate:
        column1: func_name1
        column2: func_name2

    or

    ::

        aggrgrate:
          column1:
            function: func_name
            param1: foo
            param2: baz


    Keyword Args
    ------------
    groubby : `str` or `list`
        the column(s) to group, can be a list or a string
    insert_key : `dict`
        manually insert keys in to result. This is useful when we want to add back the
        aggregated column and set them to one value. For example ``geo: global`` inserts
        the ``geo`` column with all values are "global"
    aggregate
    transform
    filter : `dict`, optinoal
        the function to run. only one of `aggregate`, `transform` and `filter` should be supplied.

    Note
    ----
    - Only one of ``aggregate``, ``transform`` or ``filter`` can be used in one procedure.
    - Any columns not mentioned in groupby or functions are dropped.

    """

    logger.info("groupby: " + ingredient.ingred_id)

    data = ingredient.get_data()
    by = options.pop('groupby')
    if 'insert_key' in options:
        insert_key = options.pop('insert_key')
    else:
        insert_key = dict()

    # only one of aggregate/transform/filter should be in options.
    assert len(list(options.keys())) == 1
    comp_type = list(options.keys())[0]
    assert comp_type in ['aggregate', 'transform', 'filter']

    if comp_type == 'aggregate':  # only aggregate should change the key of ingredient
        if isinstance(by, list):
            newkey = ','.join(by)
        else:
            newkey = by
            by = [by]
        logger.debug("changing the key to: " + str(newkey))
    else:
        newkey = ingredient.key
        by = [by]

    newdata = dict()

    # TODO: support apply function to all items?
    for k, func in options[comp_type].items():
        func = mkfunc(func)
        if comp_type == 'aggregate':
            newdata[k] = (data[k].groupby(by=by).agg({k: func})
                          .reset_index().dropna())
        if comp_type == 'transform':
            df = data[k].set_index(ingredient.key_to_list())
            levels = [df.index.names.index(x) for x in by]
            newdata[k] = (df.groupby(level=levels)[k].transform(func)
                          .reset_index().dropna())
        if comp_type == 'filter':
            df = data[k].set_index(ingredient.key_to_list())
            levels = [df.index.names.index(x) for x in by]
            newdata[k] = (df.groupby(level=levels)[k].filter(func)
                          .reset_index().dropna())
        for col, val in insert_key.items():
            newdata[k][col] = val
            newkey = newkey+','+col

    return ProcedureResult(result, newkey, data=newdata)


@debuggable
def window(ingredient: BaseIngredient, result, **options) -> ProcedureResult:
    """apply functions on a rolling window

    .. highlight:: yaml

    Procedure format:

    ::

       procedure: window
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingledient id
       options:
         window:
           column: str  # column which window is created from
           size: int or 'expanding'  # if int then rolling window, if expanding then expanding window
           min_periods: int  # as in pandas
           center: bool  # as in pandas
           aggregate: dict

    Two styles of function block are supported, and they can mix in one procedure:

    ::

       aggregate:
         col1: sum  # run rolling sum to col1
         col2: mean  # run rolling mean to col2
         col3:  # run foo to col3 with param1=baz
       function: foo
       param1: baz

    Keyword Args
    ------------
    window: dict
        window definition, see above for the dictionary format
    aggregate: dict
        aggregation functions

    Examples
    --------

    An example of rolling windows:

    .. highlight:: yaml

    ::

        procedure: window
        ingredients:
            - ingredient_to_roll
        result: new_ingredient_id
        options:
          window:
            column: year
            size: 10
            min_periods: 1
            center: false
          aggregate:
            column_to_aggregate: sum

    Notes
    -----
    Any column not mentioned in the `aggregate` block will be dropped in the returned ingredient.
    """

    logger.info('window: ' + ingredient.ingred_id)

    # reading options
    window = options.pop('window')
    aggregate = options.pop('aggregate')

    column = read_opt(window, 'column', required=True)
    size = read_opt(window, 'size', required=True)
    min_periods = read_opt(window, 'min_periods', default=0)
    center = read_opt(window, 'center', default=False)

    data = ingredient.get_data()
    newdata = dict()

    for k, func in aggregate.items():
        f = mkfunc(func)
        # keys for grouping. in multidimensional data like datapoints, we want create
        # groups before rolling. Just group all key column except the column to aggregate.
        keys = ingredient.key_to_list()
        keys.remove(column)
        df = data[k].set_index(ingredient.key_to_list())
        levels = [df.index.names.index(x) for x in keys]
        if size == 'expanding':
            newdata[k] = (df.groupby(level=levels, group_keys=False)
                          .expanding(on=column, min_periods=min_periods, center=center)
                          .agg(f).reset_index().dropna())
        else:
            # There is a bug when running rolling on with groupby in pandas.
            # see https://github.com/pandas-dev/pandas/issues/13966
            # We will implement this later when we found work around or it's fixed
            # for now, we don't use the `on` parameter in rolling.
            # FIXME: add back the `on` parameter.
            newdata[k] = (df.groupby(level=levels, group_keys=False)
                          .rolling(window=size, min_periods=min_periods, center=center)
                          .agg(f).reset_index().dropna())
    return ProcedureResult(result, ingredient.key, newdata)


@debuggable
def run_op(ingredient: BaseIngredient, result, op) -> ProcedureResult:
    """run math operation on each row of ingredient data.

    Procedure format:

    .. code-block:: yaml

       procedure: filter_item
       ingredients:  # list of ingredient id
         - ingredient_id
       result: str  # new ingledient id
       options:
         items: list  # a list of items should be in the result ingredient

    Keyword Args
    ------------
    op: dict
        a dictionary of concept_name -> function mapping

    Examples
    --------
    .. highlight:: yaml

    for exmaple, if we want to add 2 columns, col_a and col_b, to create an new column, we can
    write

    ::

        procedure: run_op
        ingredients:
          - ingredient_to_run
        result: new_ingredient_id
        options:
          op:
            new_col_name: "col_a + col_b"
    """

    assert ingredient.dtype == 'datapoints'
    logger.info("run_op: " + ingredient.ingred_id)

    data = ingredient.get_data()
    keys = ingredient.key_to_list()

    # concat all the datapoint dataframe first, and eval the ops
    to_concat = [v.set_index(keys) for v in data.values()]
    df = pd.concat(to_concat, axis=1)

    for k, v in op.items():
        res = df.eval(v).dropna()  # type(res) is Series
        res.name = k
        if k not in df.columns:
            df[k] = res
        data[k] = res.reset_index()

    if not result:
        result = ingredient.ingred_id + '-op'
    return ProcedureResult(result, ingredient.key, data=data)


@debuggable
def extract_concepts(*ingredients: List[BaseIngredient], result,
                     join=None, overwrite=None, include_keys=False) -> ProcedureResult:
    """extract concepts from other ingredients.

    .. highlight:: yaml

    Procedure format:

    ::

       procedure: extract_concepts
       ingredients:  # list of ingredient id
         - ingredient_id_1
         - ingredient_id_2
       result: str  # new ingledient id
       options:
         join:  # optional
           base: str  # base concept ingredient id
           type: {'full_outer', 'ingredients_outer'}  # default is full_outer
         overwrite:  # overwrite some concept types
           country: entity_set
           year: time
         include_keys: true  # if we should include the primaryKeys concepts

    Parameters
    ----------
    ingredients
        any numbers of ingredient that needs to extract concepts from

    Keyword Args
    ------------
    join : dict, optional
        the base ingredient to join
    overwrite : dict, optional
        overwrite concept types for some concepts
    include_keys : bool, optional
        if we shuld include the primaryKeys of the ingredients, default to false

    See Also
    --------
    :py:func:`ddf_utils.transformer.extract_concepts` : related function in transformer
    module

    Note
    ----
    - all concepts in ingredients in the ``ingredients`` parameter will be extracted
      to a new concept ingredient
    - ``join`` option is optional; if present then the ``base`` will merge with concepts
      from ``ingredients``
    - ``full_outer`` join means get the union of concepts; ``ingredients_outer`` means
      only keep concepts from ``ingredients``

    """
    if join:
        base = join['base']
        try:
            join_type = join['type']
        except KeyError:
            join_type = 'full_outer'
        concepts = base.get_data()['concepts'].set_index('concept')
    else:
        concepts = pd.DataFrame([], columns=['concept', 'concept_type']).set_index('concept')
        join_type = 'full_outer'

    new_concepts = set()

    for i in ingredients:
        data = i.get_data()
        pks = i.key_to_list()
        for k, df in data.items():
            if include_keys:
                cols = df.columns
            else:
                cols = [x for x in df.columns if x not in pks]
            for col in cols:
                new_concepts.add(col)
                if col in concepts.index:
                    continue
                if np.issubdtype(df[col].dtype, np.number):
                    concepts.ix[col, 'concept_type'] = 'measure'
                else:
                    concepts.ix[col, 'concept_type'] = 'string'
    if join_type == 'ingredients_outer':
        # ingredients_outer join: only keep concepts appears in ingredients
        concepts = concepts.ix[new_concepts]
    # overwrite some of the types
    if overwrite:
        for k, v in overwrite.items():
            concepts.ix[k, 'concept_type'] = v
    if not result:
        result = 'concepts_extracted'
    return ProcedureResult(result, 'concept', data={'concepts': concepts.reset_index()})


@debuggable
def trend_bridge(ingredient: BaseIngredient, bridge_start, bridge_end, bridge_length, bridge_on,
                 result, target_column=None) -> ProcedureResult:
    """run trend bridge on ingredients

    .. highlight:: yaml

    Procedure format:

    ::

      procedure: trend_bridge
      ingredients:
        - data_ingredient                 # optional, if not set defaults to empty ingredient
      result: data_bridged
      options:
        bridge_start:
            ingredient: old_data_ingredient # optional, if not set then assume it's the input ingredient
            column: concept_old_data
        bridge_end:
            ingredient: new_data_ingredient # optional, if not set then assume it's the input ingredient
            column: concept_new_data
        bridge_length: 5                  # steps in time. If year, years, if days, days.
        bridge_on: time                   # the index column to build the bridge with
        target_column: concept_in_result  # overwrites if exists. creates if not exists. default to bridge_end.column

    Parameters
    ----------
    ingredient : BaseIngredient
        The input ingredient. The bridged result will be merged in to this ingredient. If this is
        None, then the only the bridged result will be returned
    bridge_start : dict
        Describe the start of bridge
    bridge_end : dict
        Describe the end of bridge
    bridge_length : int
        The size of bridge
    bridge_on : `str`
        The column to bridge
    result : `str`
        The output ingredient id

    Keyword Args
    ------------
    target_column : `str`, optional
        The column name of the bridge result. default to `bridge_end.column`

    See Also
    --------
    :py:func:`ddf_utils.transformer.trend_bridge` : related function in transformer module
    """
    from ..transformer import trend_bridge as tb

    # check paramaters
    if ingredient is None:
        assert 'ingredient' in bridge_start.keys()
        assert 'ingredient' in bridge_end.keys()

    # get data for start and end
    if 'ingredient' in bridge_start.keys():
        start = bridge_start['ingredient']
    else:
        start = ingredient
    if 'ingredient' in bridge_end.keys():
        end = bridge_end['ingredient']
    else:
        end = ingredient

    assert start.dtype == 'datapoints'
    assert end.dtype == 'datapoints'

    if target_column is None:
        target_column = bridge_start['column']

    # get the column to group. Because datapoints are multidimensional, but we only
    # bridge them in one column, so we should group other columns.
    assert set(start.key_to_list()) == set(end.key_to_list())

    keys = start.key_to_list()
    keys.remove(bridge_on)

    start_group = start.get_data()[bridge_start['column']].set_index(bridge_on).groupby(keys)
    end_group = end.get_data()[bridge_end['column']].set_index(bridge_on).groupby(keys)

    # calculate trend bridge on each group
    res_grouped = []
    for g, df in start_group:
        gstart = df.copy()
        try:
            gend = end_group.get_group(g)
        except KeyError:  # no new data available for this group
            logger.warning("no data for bridge end: " + g)
            bridged = gstart[bridge_start['column']]
        else:
            bridged = tb(gstart[bridge_start['column']], gend[bridge_end['column']], bridge_length)

        res_grouped.append((g, bridged))

    # combine groups to dataframe
    res = []
    for g, v in res_grouped:
        v.name = target_column
        v = v.reset_index()
        if len(keys) == 1:
            assert isinstance(g, str)
            v[keys[0]] = g
        else:
            assert isinstance(g, list)
            for i, k in enumerate(keys):
                v[k] = g[i]
        res.append(v)
    result_data = pd.concat(res, ignore_index=True)

    if ingredient is not None:
        merged = _merge_two(ingredient.get_data(), {target_column: result_data},
                            start.key_to_list(), 'datapoints')
        return ProcedureResult(result, start.key, merged)
    else:
        return ProcedureResult(result, start.key, {target_column: result_data})


@debuggable
def merge_entity(ingredient: BaseIngredient, dictionary,
                 target_column, result, merged='drop'):
    """merge entities"""
    from ..transformer import merge_keys

    data = ingredient.get_data()

    res_data = dict()
    for k, df in data.items():
        res_data[k] = merge_keys(df.set_index(ingredient.key_to_list()),
                                 dictionary, merged).reset_index()

    return ProcedureResult(result, ingredient.key, res_data)


@debuggable
def split_entity(ingredient: BaseIngredient, dictionary,
                 target_column, result, splitted='drop'):
    """split entities"""
    from ..transformer import split_keys

    data = ingredient.get_data()

    res_data = dict()
    for k, df in data.items():
        res_data[k] = split_keys(df.set_index(ingredient.key_to_list()),
                                 target_column, dictionary, splitted).reset_index()

    return ProcedureResult(result, ingredient.key, res_data)
